"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_transit_gateway
hub.exec.boto3.client.ec2.delete_transit_gateway
hub.exec.boto3.client.ec2.describe_transit_gateways
hub.exec.boto3.client.ec2.modify_transit_gateway
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
TREQ = {
    "present": {
        "require": ["aws.ec2.subnet.present"],
        "require": ["aws.ec2.transit_gateway.present"],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    transit_gateway: str,
    vpc_id: str,
    subnet_ids: List,
    resource_id: str = None,
    options: dict = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Attaches the specified VPC to the specified transit gateway. If you attach a VPC with a CIDR range that overlaps
    the CIDR range of a VPC that is already attached, the new VPC CIDR range is not propagated to the default
    propagation route table. To send VPC traffic to an attached transit gateway, add a route to the VPC route table
    using CreateRoute.

    Args:
        name(Text): Name of the resource.
        transit_gateway(Text): The ID of the transit gateway.
        vpc_id(Text): The ID of the VPC.
        subnet_ids(List): The IDs of one or more subnets. You can specify only one subnet per Availability Zone.
                       You must specify at least one subnet, but we recommend that you specify two subnets for better availability.
                       The transit gateway uses one IP address from each specified subnet.
             * (Text)
        resource_id(Text, optional): AWS Transit gateway VPC attachment id to identify the resource
        options(dictionary, optional): The VPC attachment options.
             *  DnsSupport (Text) -- Enable or disable DNS support. Enabled by default.,
             *  Ipv6Support (Text) -- Enable or disable IPv6 support. The default is disable.,
             *  ApplianceModeSupport (Text) -- Enable or disable support for appliance mode. If enabled, a traffic flow between a source and destination uses the same Availability Zone for the VPC attachment for the lifetime of that flow. The default is disable .
        tags(List, optional): The tags to apply to the VPC attachment.
            * Key (Text) -- The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
            * Value (Text) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.

    Request Syntax:
        [transit-gateway-resource-id]:
          aws.ec2.transit_gateway_vpc_attachment.present:
            - name: 'string'
            - resource_id: 'string'
            - transit_gateway: 'string'
            - vpc_id: 'string'
            - subnet_ids:
               - 'string'
            - options:
                ApplianceModeSupport: 'string'
                DnsSupport: 'string'
                Ipv6Support: 'string'
            - tags:
                - Key: 'string'
                  Value: 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            tgw-attach-0871e9c72becf0710:
                aws.ec2.transit_gateway_vpc_attachment.present:
                - name: tgw-attach-0871e9c72becf0710
                - resource_id: tgw-attach-0871e9c72becf0710
                - transit_gateway: tgw-02994a8dda824c337
                - vpc_id: vpc-0afa0d5fe3fc2785c
                - subnet_ids:
                   - subnet-07f91b9ebd252be49
                - options:
                    ApplianceModeSupport: disable
                    DnsSupport: enable
                    Ipv6Support: disable
                - tags:
                    - Key: Name
                      Value: test-transit-gateway-attachment

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.get_transit_gateway_vpc_attachment_by_id(
            ctx, resource_id
        )
    if before and before["ret"]:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
            before["ret"], name
        )

    try:
        if before and before["ret"]:
            plan_state = copy.deepcopy(result["old_state"])
            if subnet_ids is not None:
                update_ret = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.update_transit_gateway_vpc_attachment(
                    ctx=ctx,
                    transit_gateway_vpc_attachment_id=resource_id,
                    old_subnets=before["ret"].get("SubnetIds"),
                    new_subnets=subnet_ids,
                    old_options=before["ret"].get("Options"),
                    new_options=options,
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or update_ret["result"]
                if ctx.get("test", False) and update_ret["ret"]:
                    if subnet_ids:
                        plan_state["subnet_ids"] = update_ret["ret"].get("subnet_ids")
                    if options:
                        plan_state["options"] = update_ret["ret"].get("options")
            if result["result"] and (tags is not None):
                # Update tags
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx=ctx,
                    resource_id=resource_id,
                    old_tags=before.get("Tags"),
                    new_tags=tags,
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = update_ret["result"]
                resource_updated = resource_updated or update_ret["result"]
                if ctx.get("test", False) and update_ret["ret"]:
                    plan_state["tags"] = tags

            if result["result"]:
                result["comment"] = result["comment"] + (
                    f"'Updated transit gateway {name}'",
                )

        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "transit_gateway": transit_gateway,
                        "vpc_id": vpc_id,
                        "subnet_ids": subnet_ids,
                        "options": options,
                        "tags": tags,
                    },
                )
                result["comment"] = result["comment"] + (
                    f"Would create aws.ec2.transit_gateway_vpc_attachment {name}",
                )
                return result
            ret = await hub.exec.boto3.client.ec2.create_transit_gateway_vpc_attachment(
                ctx,
                TransitGatewayId=transit_gateway,
                VpcId=vpc_id,
                SubnetIds=subnet_ids,
                Options=options,
                TagSpecifications=[
                    {"ResourceType": "transit-gateway-attachment", "Tags": tags}
                ]
                if tags
                else None,
            )

            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = ret["result"]
                return result

            result["comment"] = result["comment"] + (f"Created '{name}'",)
            resource_updated = resource_updated or ret["result"]
            resource_id = ret["ret"]["TransitGatewayVpcAttachment"][
                "TransitGatewayAttachmentId"
            ]

    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif resource_updated:
            after = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.get_transit_gateway_vpc_attachment_by_id(
                ctx, resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
                after["ret"], name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified VPC attachment.

    Args:
        name(Text): Name of the resource.
        resource_id(Text): AWS Transit gateway VPC attachment ID to identify the resource

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            aws.ec2.transit_gateway_vpc_attachment.absent:
                - name: tgw-attach-0871e9c72becf0710
                - resource_id: tgw-attach-0871e9c72becf0710
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.get_transit_gateway_vpc_attachment_by_id(
        ctx, resource_id
    )
    if not before or not before["ret"]:
        result["comment"] = (f"'{name}' already absent",)
        return result
    before = before["ret"]
    result[
        "old_state"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(before, name)
    if before["State"] and (
        before["State"] == "deleted" or before["State"] == "deleting"
    ):
        result["comment"] = result["comment"] + (
            f"'{name}' is already {before['State']}",
        )
    elif ctx.get("test", False):
        result["comment"] = result["comment"] + (
            f"Would delete aws.ec2.transit_gateway_vpc_attachment {name}",
        )
        return result
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.delete_transit_gateway_vpc_attachment(
                ctx, TransitGatewayAttachmentId=resource_id
            )
            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = ret["result"]
                return result
            result["comment"] = result["comment"] + (f"Deleted '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more VPC attachments. By default, all VPC attachments are described. Alternatively, you can
    filter the results.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.transit_gateway_vpc_attachment
    """

    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_transit_gateway_vpc_attachments(ctx)
    if not ret["result"]:
        hub.log.debug(
            f"Could not describe transit_gateway_vpc_attachments {ret['comment']}"
        )
        return {}
    for transit_gateway_vpc_attachment in ret["ret"]["TransitGatewayVpcAttachments"]:
        resource_id = transit_gateway_vpc_attachment.get("TransitGatewayAttachmentId")
        translated_resource = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
                transit_gateway_vpc_attachment, resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.transit_gateway_vpc_attachment.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.organizations.deregister_delegated_administrator
hub.exec.boto3.client.organizations.list_delegated_administrators
hub.exec.boto3.client.organizations.register_delegated_administrator
"""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
SERVICE = "organizations"


TREQ = {
    "absent": {
        "require": [
            "aws.organizations.policy_attachment.absent",
            "aws.organizations.policy.absent",
        ],
    },
    "present": {
        "require": [
            "aws.organizations.organization.present",
            "aws.organizations.organization_unit.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    email: str,
    account_name: str,
    role_name: str,
    iam_user_access_to_billing: str = "ALLOW",
    parent_id: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an AWS account that is automatically a member of the organization whose credentials made the request.
    This is an asynchronous request that AWS performs in the background. Because CreateAccount operates asynchronously,
    it can return a successful completion message even though account initialization might still be in progress.
     You might need to wait a few minutes before you can successfully access the account

    Args:
        hub:
        ctx:
        name(Text): A name or ID to identify the resource.
        Email(str, Required): The email address of the owner to assign to the new member account. This email
                              address must not already be associated with another AWS account
        AccountName(str,Required): The friendly name of the member account.
        RoleName(str,Optional): The name of an IAM role that AWS Organizations automatically preconfigures in the
                                new member account. This role trusts the management account, allowing users in the
                                management account to assume the role, as permitted by the management account
                                administrator. The role has administrator permissions in the new member account.
                                If you don't specify this parameter, the role name defaults to OrganizationAccountAccessRole.
        IamUserAccessToBilling(str,Optional,Default:'ALLOW'): If set to ALLOW , the new account enables IAM users to access account
                                                              billing information if they have the required permissions. If set to DENY ,
                                                              only the root user of the new account can access account billing information.
        ParentId(str,Optional): Parent Organizational Unit ID or Root ID for the account. Defaults to the Organization default Root ID
        Tags(list,Optional) : A list of tags that you want to attach to the newly created account

    Request Syntax:
        [account-id]:
          aws.organizations.account.present:
          - email: 'string'
          - account_name: 'string'
          - role_name: 'string'
          - iam_user_access_to_billing: 'string'
          - parent_id: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            new-account_id:
                aws.organizations.account.present:
                    - email: xyz@email.com
                    - account_name: name_of_new_account
                    - role_name: role_name1
                    - iam_user_access_to_billing: ALLOW
                    - parent_id : ou108811
                    - tags:
                        - Key: test-key
                          Value: test-value
                        - Key: test-key-1
                          Value: test-key-1
    """

    result = dict(comment="", name=name, result=True, old_state=None, new_state=None)

    before = await hub.exec.boto3.client.organizations.describe_account(
        ctx, AccountId=name
    )

    account_id = None

    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.organizations.account {name}."
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.oranizations.organization {name}."
            result["result"] = True
        return result

    update_tag = False
    update_parent = False

    if before:
        # Account exists , update
        before["ret"].pop("ResponseMetadata", None)
        result["old_state"] = before["ret"]["Account"]

        account_id = name

        result["comment"] = f"aws.organizations.account {name} already exists . "

        try:

            # we need to list parents to check if move is required in case the new_parent_id is not equal to current_parent_id
            parents = await hub.exec.boto3.client.organizations.list_parents(
                ctx, ChildId=account_id
            )

            if parents:
                current_parent_id = parents["ret"]["Parents"][0]["Id"]

                result["old_state"]["ParentId"] = current_parent_id

                if parent_id is not None and current_parent_id != parent_id:
                    await move_account(hub, ctx, account_id, result, parent_id)
                    update_parent = True

            old_tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
                ctx, ResourceId=account_id
            )

            if old_tags:
                before["ret"]["Account"]["Tags"] = old_tags["ret"]["Tags"]
                result["old_state"]["Tags"] = old_tags["ret"]["Tags"]

            if tags is not None and old_tags["result"]:

                update_tags_ret = (
                    await hub.exec.aws.organizations.organization.update_tags(
                        ctx, account_id, old_tags["ret"].get("Tags", []), tags
                    )
                )
                if not update_tags_ret["result"]:
                    result[
                        "comment"
                    ] = f" {result['comment']} , {str(update_tags_ret['comment'])}"
                    result["result"] = False
                    return result

                old_comment = result["comment"]

                if update_tags_ret["ret"] is not None:
                    result[
                        "comment"
                    ] = f"'{old_comment}'. Updated tags on aws.organizations.account '{name}'."
                    update_tag = True

            result["result"] = True

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    else:
        # Account not present , create
        hub.log.debug("Creating member account in the organization")

        try:
            create_account_ret = (
                await hub.exec.boto3.client.organizations.create_account(
                    ctx,
                    Email=email,
                    AccountName=account_name,
                    RoleName=role_name,
                    IamUserAccessToBilling=iam_user_access_to_billing,
                    Tags=tags,
                )
            )

            result["result"] = create_account_ret["result"]
            if not result["result"]:
                result["comment"] = create_account_ret["comment"]
                return result

            account_status_id = create_account_ret["ret"]["CreateAccountStatus"]["Id"]
            account_id = create_account_ret["ret"]["CreateAccountStatus"].get(
                "AccountId", None
            )

            # Call a custom waiter to wait on account's creation.

            acceptors = {
                "SUCCEEDED": "success",
                "FAILED": "failure",
                "IN_PROGRESS": "retry",
            }
            account_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                name="AccountCreated",
                operation="DescribeCreateAccountStatus",
                argument="CreateAccountStatus.State",
                acceptors=acceptors,
                client=hub.tool.boto3.client.get_client(ctx, SERVICE),
            )
            await hub.tool.boto3.client.wait(
                ctx,
                SERVICE,
                "AccountCreated",
                account_waiter,
                CreateAccountRequestId=account_status_id,
            )

            result["comment"] = f"Created aws.organizations.account {name}."

            if account_id is None:
                create_account_status = await hub.exec.boto3.client.organizations.describe_create_account_status(
                    ctx, CreateAccountRequestId=account_status_id
                )
                if not create_account_status:
                    result[
                        "comment"
                    ] = f"{result['comment']} . Could not find account_id for {account_status_id} in aws.organizations.account {name}"
                else:
                    account_id = create_account_status["ret"]["CreateAccountStatus"][
                        "AccountId"
                    ]

            if account_id is not None and parent_id is not None:

                parents = await hub.exec.boto3.client.organizations.list_parents(
                    ctx, ChildId=account_id
                )

                if parents:
                    current_parent_id = parents["ret"]["Parents"][0]["Id"]

                    if current_parent_id != parent_id:
                        await move_account(hub, ctx, account_id, result, parent_id)
            else:
                result[
                    "comment"
                ] = f"{result['comment']} . Could not update parent for aws.organizations.account {name}"
                result["result"] = False
                return result

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    if not before or update_parent or update_tag:
        try:
            after = await hub.exec.boto3.client.organizations.describe_account(
                ctx, AccountId=account_id
            )
            if after and after.get("ret"):
                after["ret"].pop("ResponseMetadata", None)
                result["new_state"] = after["ret"]["Account"]

                parents = await hub.exec.boto3.client.organizations.list_parents(
                    ctx, ChildId=account_id
                )
                if parents:
                    current_parent_id = parents["ret"]["Parents"][0]["Id"]
                    result["new_state"]["ParentId"] = current_parent_id

                new_tags = (
                    await hub.exec.boto3.client.organizations.list_tags_for_resource(
                        ctx, ResourceId=account_id
                    )
                )
                if new_tags:
                    result["new_state"]["Tags"] = new_tags["ret"]["Tags"]

        except Exception as e:
            result["comment"] = str(e)
            result["result"] = False

    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Removes the specified account from the organization.The removed account becomes a standalone
    account that isn't a member of any organization. It's no longer subject to any policies and
    is responsible for its own bill payments. The organization's management account is no longer
    charged for any expenses accrued by the member account after it's removed from the organization.
    This operation can be called only from the organization's management account. Member accounts
    can remove themselves with LeaveOrganization instead.

    Args:
        name(Text): The unique identifier (ID) of the member account that you want
                        to remove from the organization

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            [account_id]:
              aws.organizations.organization.absent:
                - name: value
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)

    account_id = name

    before = await hub.exec.boto3.client.organizations.describe_account(
        ctx, AccountId=account_id
    )

    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would delete aws.organizations.account {name}."
            result["result"] = True
        else:
            result["comment"] = f"aws.organizations.account {name} already absent"
            result["result"] = True
        return result

    if not before:
        result["comment"] = f"aws.organizations.account {name} already absent"
        result["result"] = True
    else:
        before["ret"].pop("ResponseMetadata", None)

        result["old_state"] = before["ret"]["Account"]
        try:
            ret = await hub.exec.boto3.client.organizations.remove_account_from_organization(
                ctx, AccountId=account_id
            )

            result["result"] = ret["result"]

            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"aws.organizations.account {name} deleted."
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Lists all the accounts in the organization. To request only the accounts in a specified root or organizational
    unit (OU), use the ListAccountsForParent operation instead.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.organizations.account
    """

    result = {}

    describe_ret = await hub.exec.boto3.client.organizations.list_accounts(ctx)
    if not describe_ret:
        hub.log.debug(f"Could not describe account {describe_ret['comment']}")
        return {}

    describe_parameters = OrderedDict(
        {
            "Name": "account_name",
            "Email": "email",
            "ParentId": "parent_id",
            "Tags": "tags",
        }
    )

    accounts = describe_ret["ret"]["Accounts"]

    for account in accounts:
        translated_resource = []
        if account.get("Name") is not None:
            translated_resource.append({"account_name": account.get("Name")})
        if account.get("Email") is not None:
            translated_resource.append({"email": account.get("Email")})

        parent = await hub.exec.boto3.client.organizations.list_parents(
            ctx, ChildId=account["Id"]
        )

        if parent and parent["ret"].get("Parents"):
            translated_resource.append({"parent_id": parent["ret"]["Parents"][0]["Id"]})

        tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=account["Id"]
        )

        if tags and tags["ret"].get("Tags"):
            translated_resource.append({"tags": tags["ret"].get("Tags")})

        result[account["Id"]] = {
            "aws.organizations.account.present": translated_resource
        }
    return result


async def move_account(
    hub, ctx, account_id, result: Dict[str, Any], destination_parent_id
):
    if destination_parent_id is not None:

        parents = await hub.exec.boto3.client.organizations.list_parents(
            ctx, ChildId=account_id
        )

        if parents:
            parent_id = parents["ret"]["Parents"][0]["Id"]

            if parent_id == destination_parent_id:
                result[
                    "comment"
                ] = f"{result['comment']} . aws.organizations.account {account_id} already at {destination_parent_id}"
                return
            else:

                account_move_ret = (
                    await hub.exec.boto3.client.organizations.move_account(
                        ctx,
                        AccountId=account_id,
                        SourceParentId=parent_id,
                        DestinationParentId=destination_parent_id,
                    )
                )
                if not account_move_ret:
                    result[
                        "comment"
                    ] = f"{result['comment']} . Could not move aws.organizations.account {account_id} to {destination_parent_id}"

                    result["result"] = False

                    return
                result[
                    "comment"
                ] = f"{result['comment']} Successfully moved aws.organizations.account {account_id} under {destination_parent_id}"



from typing import Dict
from typing import List

ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS = {
    "DestinationCidrBlock",
    "DestinationIpv6CidrBlock",
    "DestinationPrefixListId",
}

ALLOWED_ROUTE_TARGET_GATEWAYS = {
    "VpcEndpointId",
    "EgressOnlyInternetGatewayId",
    "GatewayId",
    "InstanceId",
    "NatGatewayId",
    "TransitGatewayId",
    "LocalGatewayId",
    "CarrierGatewayId",
    "NetworkInterfaceId",
    "RouteTableId",
    "VpcPeeringConnectionId",
    "CoreNetworkArn",
}

ALLOWED_DEFAULT_ROUTE_TARGET_GATEWAYS = {
    "VpcEndpointId",
    "GatewayId",
    "InstanceId",
    "NetworkInterfaceId",
}


def get_route_table_routes_modifications(
    hub, old_routes: List[Dict], new_routes: List[Dict]
):
    """
    Given old routes i.e, the routes of current route table and new_routes
    i.e, the routes we want to modify to, this function will compare differences between
    old and new routes and return routes to be added and removed

    Args:
        hub:
        old_routes(List):  routes of existing route table
        new_routes(List): new routes to update

    Returns:
        Dict[str, List]

    """
    result = dict(
        comment="",
        routes_to_add=list(),
        routes_to_delete=list(),
        routes_to_replace=list(),
        result=True,
    )
    routes_to_add = list()
    routes_to_delete = list()
    routes_to_replace = list()
    old_routes_map = dict()
    default_destination_cidr_block: str = ""
    # Creates a map of destination cidr block -> route to easily retrieve further
    for route in old_routes:
        if route.get("State") == "active":
            # Route table route allows three types of destinations
            # 1) DestinationCidrBlock,
            # 2) DestinationIpv6CidrBlock,
            # 3) DestinationPrefixListId,
            # We need to determine which type of  destination is used
            destination_cidr_key = get_common_key_in_dict(
                route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
            )
            old_routes_map[route.get(destination_cidr_key)] = route
            if route.get("Origin") == "CreateRouteTable":
                default_destination_cidr_block = route.get(destination_cidr_key)
    new_destination_cidr_blocks = list()
    for new_route in new_routes:
        # Loop through new routes and determine which destination is used
        destination_cidr_key = get_common_key_in_dict(
            new_route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
        )
        destination_cidr_block = new_route.get(destination_cidr_key)

        # checking for duplicate destination cidr blocks. There can be only unique destination cidr block
        if destination_cidr_block in new_destination_cidr_blocks:
            result["result"] = False
            result[
                "comment"
            ] = "Duplicate destination cidr blocks are not allowed, they should be unique."
            return result
        else:
            new_destination_cidr_blocks.append(destination_cidr_block)
        allowed_targets = (
            ALLOWED_DEFAULT_ROUTE_TARGET_GATEWAYS
            if destination_cidr_block == default_destination_cidr_block
            else ALLOWED_ROUTE_TARGET_GATEWAYS
        )
        new_target_key = get_common_key_in_dict(new_route, allowed_targets)
        if new_target_key is None or destination_cidr_key is None:
            result["result"] = False
            result[
                "comment"
            ] = "There should be at least one valid destination cidr block and one target"
            return result
        new_target_value = new_route.get(new_target_key)
        # Destination cidr is unique for a route. After retrieving destination cidr check in old routes
        # if the new route cidr block is not present in old routes we associate it to route table
        # if the destination cidr is already present in old routes we need to check if there is any changes in target
        # The allowed targets can be any one of Internet gateway, Virtual private gateway and many more.
        # If the target key is modified or value is modified we need to delete old route and create
        # a new route with updated values
        # example the target might be changed from an Internet gateway to Transit gateway the key changes
        # So we need to delete Internet gateway route and create Transit gateway route
        if destination_cidr_block not in old_routes_map:
            routes_to_add.append(
                {
                    destination_cidr_key: destination_cidr_block,
                    new_target_key: new_target_value,
                }
            )
        else:
            old_route = old_routes_map.get(destination_cidr_block)
            old_target_key = get_common_key_in_dict(
                old_route, ALLOWED_ROUTE_TARGET_GATEWAYS
            )
            old_target_value = old_route.get(old_target_key)
            if old_target_key != new_target_key or old_target_value != new_target_value:
                routes_to_replace.append(
                    {
                        destination_cidr_key: destination_cidr_block,
                        new_target_key: new_target_value,
                    }
                )
            del old_routes_map[destination_cidr_block]

    # delete the remaining routes which are still present in old routes and not in new routes
    for old_route in old_routes_map.values():
        destination_cidr_key = get_common_key_in_dict(
            old_route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
        )
        destination_cidr_block = old_route.get(destination_cidr_key)
        # we cannot delete default destination cidr block
        if destination_cidr_block == default_destination_cidr_block:
            result["result"] = False
            result[
                "comment"
            ] = "We cannot delete default destination cidr block. we can only change the target."
            return result
        routes_to_delete.append({destination_cidr_key: destination_cidr_block})

    result["routes_to_add"] = routes_to_add
    result["routes_to_delete"] = routes_to_delete
    result["routes_to_replace"] = routes_to_replace
    return result


def get_route_table_associations_modifications(
    hub, old_associations: List[Dict], new_associations: List[Dict]
):
    """
    Given old associations i.e, the associations of current association table and new_associations
    i.e, the associations we want to modify to, this function will compare differences between
    old and new associations and return associations to be added and removed

    Args:
        hub:
        old_associations(List):  associations of existing association table
        new_associations(List): new associations to update

    Returns:
        Dict[str, List]

    """
    result = dict(
        comment="",
        associations_to_add=list(),
        associations_to_delete=list(),
        result=True,
    )
    associations_to_add = list()
    associations_to_delete = list()
    old_associations_map = dict()
    # Creates a map of Route table association ID -> association to easily retrieve further
    for association in old_associations:
        if (
            association.get("AssociationState")
            and association.get("AssociationState").get("State") == "associated"
            and association.get("Main") is not True
        ):
            old_associations_map[
                association.get("RouteTableAssociationId")
            ] = association
    new_route_table_association_ids = list()
    for new_association in new_associations:
        # Loop through new associations and determine which target is used
        route_table_association_id = new_association.get("RouteTableAssociationId")
        new_target_key = None

        if "SubnetId" in new_association or "GatewayId" in new_association:
            new_target_key = (
                "SubnetId" if "SubnetId" in new_association else "GatewayId"
            )

        if new_target_key is None:
            result["result"] = False
            result[
                "comment"
            ] = f"There should be at least one valid Subnet or GatewayId. {new_association}"
            return result
        new_target_value = new_association.get(new_target_key)
        # checking for duplicate route table association ids. There can be only unique destination cidr block
        if route_table_association_id is not None:
            if route_table_association_id in new_route_table_association_ids:
                result["result"] = False
                result[
                    "comment"
                ] = "Duplicate route table associations are not allowed, they should be unique."
                return result
            else:
                new_route_table_association_ids.append(route_table_association_id)

        # Destination cidr is unique for an association. After retrieving destination cidr check in old associations
        # if the new association cidr block is not present in old associations we associate it to association table
        # if the destination cidr is already present in old associations we need to check
        # if there is any changes in target
        # The allowed targets can be any one of Internet gateway, Virtual private gateway and many more.
        # If the target key is modified or value is modified we need to delete old association and create
        # a new association with updated values
        # example the target might be changed from an Internet gateway to Transit gateway the key changes
        # So we need to delete Internet gateway association and create Transit gateway association
        if (
            route_table_association_id is None
            or route_table_association_id not in old_associations_map
        ):
            associations_to_add.append({new_target_key: new_target_value})
        else:
            old_association = old_associations_map.get(route_table_association_id)
            old_target_key = None
            if "SubnetId" in old_association or "GatewayId" in old_association:
                old_target_key = (
                    "SubnetId" if "SubnetId" in new_association else "GatewayId"
                )
            old_target_value = old_association.get(old_target_key)
            if old_target_key != new_target_key or old_target_value != new_target_value:
                associations_to_add.append({new_target_key: new_target_value})
                associations_to_delete.append(
                    {"AssociationId": old_association.get("RouteTableAssociationId")}
                )
            del old_associations_map[route_table_association_id]

    # delete the remaining associations which are still present in old associations and not in new associations
    for old_association in old_associations_map.values():
        associations_to_delete.append(
            {"AssociationId": old_association.get("RouteTableAssociationId")}
        )

    result["associations_to_add"] = associations_to_add
    result["associations_to_delete"] = associations_to_delete
    return result


def get_route_table_propagating_vgws_modifications(
    hub, old_propagating_vgws: List[Dict], new_propagating_vgws: List[Dict]
):
    result = dict(
        comment="",
        propagating_vgws_to_add=list(),
        propagating_vgws_to_delete=list(),
        result=True,
    )
    old_propagating_vgw_ids = []
    new_propagating_vgw_ids = []
    if old_propagating_vgws:
        for old_propagating_vgw in old_propagating_vgws:
            old_propagating_vgw_ids.append(old_propagating_vgw.get("GatewayId"))
    if new_propagating_vgws:
        for new_propagating_vgw in new_propagating_vgws:
            new_propagating_vgw_ids.append(new_propagating_vgw.get("GatewayId"))

    result["propagating_vgws_to_add"] = list(
        set(new_propagating_vgw_ids).difference(old_propagating_vgw_ids)
    )
    result["propagating_vgws_to_delete"] = list(
        set(old_propagating_vgw_ids).difference(new_propagating_vgw_ids)
    )
    return result


def get_common_key_in_dict(dict1: dict, target_keys_list: set):
    common_keys = set(dict1.keys()).intersection(target_keys_list)
    return common_keys.pop() if common_keys else None



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.create_policy
hub.exec.boto3.client.iam.delete_policy
hub.exec.boto3.client.iam.get_policy
hub.exec.boto3.client.iam.list_policies
hub.exec.boto3.client.iam.tag_policy
hub.exec.boto3.client.iam.untag_policy
resource = hub.tool.boto3.resource.create(ctx, "iam", "Policy", name)
hub.tool.boto3.resource.exec(resource, attach_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_role, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_user, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_version, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_role, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_user, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    policy_name: str,
    policy_document: Dict or str,
    resource_id: str = None,
    path: str = "/",
    description: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Create or update a managed policy for your Amazon Web Services account. When creating a policy, this operation
     creates a policy version with a version identifier of v1 and sets v1 as the policy's default version. When updating
      a policy, this operation creates a new policy version, sets the new policy to the default version, and
       deletes the old policy. As a best practice, you can validate your IAM policies. To learn more, see Validating
        IAM policies in AWS's IAM User Guide. For more information about managed policies in general,
         see Managed policies and inline policies in AWS's IAM User Guide.

    Args:
        name(Text): A name, or the Amazon Resource Name (ARN) to identify the resource.
        policy_name(Text): The friendly name of the policy. Policy names must be unique within the account.
         They are not distinguished by case.
        policy_document(Dict or Text): The JSON policy document that you want to use as the content for the new policy.
         You must provide policies in JSON format in IAM. However, for CloudFormation templates formatted in YAML,
          you can provide the policy in JSON or YAML format. CloudFormation always converts a YAML policy to JSON format
           before submitting it to IAM.
        resource_id(Text, Optional): The ID of the resource.
        path(Text, Optional): The path for the policy. Defaults to a slash (/)
        description(Text, Optional): A friendly description of the policy.
        tags(List, Optional): A list of tags that you want to attach to the new IAM customer managed policy.
         Each tag consists of a key name and an associated value.

    Request Syntax:
        [policy-resource-id]:
          aws.iam.policy.present:
          - policy_name: 'string'
          - path: 'string'
          - policy_document: 'dict' or 'string'
          - description: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            arn:aws:iam::53721234567:policy/idem-policy-automation:
              aws.iam.policy.present:
                - policy_name: idem-policy-automation
                - path: /
                - policy_document: '{"Statement": [{"Action": ["ec2:CreateSubnet"], "Effect": "Allow", "Resource": "*"}], "Version": "2012-10-17"}'
                - description: "Idem IAM policy description example"
                - tags:
                  - Key: Name
                    Value: Idem-test-policy
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(
        ctx, "iam", "Policy", resource_id if resource_id else name
    )
    before = await hub.tool.boto3.resource.describe(resource)

    # Standardise on the json format
    policy_document = hub.tool.aws.state_comparison_utils.standardise_json(
        policy_document
    )

    update_ret = None
    if before:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
                ctx, before
            )
            result["comment"] = (f"'{name}' already exists",)
            plan_state = copy.deepcopy(result["old_state"])
            # Update policy document if needed
            if (
                result["old_state"] is not None
                and result["old_state"]["policy_document"]
                and not hub.tool.aws.state_comparison_utils.is_json_identical(
                    result["old_state"]["policy_document"], policy_document
                )
            ):
                if ctx.get("test", False):
                    plan_state["policy_document"] = policy_document
                    result["comment"] = result["comment"] + (
                        f"Would update policy document.",
                    )
                else:
                    update_ret = await hub.exec.aws.iam.policy.update_policy(
                        ctx=ctx,
                        policy_arn=resource_id,
                        policy_version_id=result["old_state"].get("default_version_id"),
                        new_policy_document=policy_document,
                    )
                    if not update_ret["result"]:
                        result["comment"] = result["comment"] + update_ret["comment"]
                        result["result"] = False

            # Update policy tags
            if (
                result["old_state"] is not None
                and tags is not None
                and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    tags, result["old_state"].get("tags", None)
                )
            ):
                update_ret = await hub.exec.aws.iam.policy.update_policy_tags(
                    ctx=ctx,
                    police_arn=resource_id,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = False
                if ctx.get("test", False) and update_ret["ret"] is not None:
                    plan_state["tags"] = update_ret["ret"].get("tags")
                    result["comment"] = result["comment"] + ("Would update tags.",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

        if update_ret is not None and result["result"] and not ctx.get("test", False):
            result["comment"] = (f"'Updated policy {name}'",)
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "policy_name": name,
                        "path": path,
                        "policy_document": policy_document,
                        "description": description,
                        "tags": tags,
                    },
                )
                result["comment"] = (f"Would create aws.iam.policy {name}",)
                return result

            ret = await hub.exec.boto3.client.iam.create_policy(
                ctx,
                PolicyName=policy_name,
                Path=path,
                PolicyDocument=policy_document,
                Description=description,
                Tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = (f"Created aws.iam.policy '{name}'",)
            resource_id = ret["ret"]["Policy"]["Arn"]
            resource = hub.tool.boto3.resource.create(ctx, "iam", "Policy", resource_id)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif before and update_ret is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
        else:
            after = await hub.tool.boto3.resource.describe(resource)
            result[
                "new_state"
            ] = await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
                ctx, after
            )
    except Exception as e:
        result["comment"] = (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified managed policy. Before you can delete a managed policy, you must first detach the policy
    from all users, groups, and roles that it is attached to. In addition, you must delete all the policy's
    versions. The following steps describe the process for deleting a managed policy:   Detach the policy from all
    users, groups, and roles that the policy is attached to, using DetachUserPolicy, DetachGroupPolicy, or
    DetachRolePolicy. To list all the users, groups, and roles that a policy is attached to, use
    ListEntitiesForPolicy.   Delete all versions of the policy using DeletePolicyVersion. To list the policy's
    versions, use ListPolicyVersions. You cannot use DeletePolicyVersion to delete the version that is marked as the
    default version. You delete the policy's default version in the next step of the process.   Delete the policy
    (this automatically deletes the policy's default version) using this operation.   For information about managed
    policies, see Managed policies and inline policies in the IAM User Guide.

    Args:
        name(Text): A name
        resource_id(Text, Optional): Amazon Resource Name (ARN) to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.iam.policy.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(
        ctx, "iam", "Policy", resource_id if resource_id else name
    )
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = (f"aws.iam.policy '{name}' already absent",)
    else:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
                ctx, raw_resource=before
            )

            if ctx.get("test", False):
                result["comment"] = (f"Would delete aws.iam.policy '{name}'",)
                return result

            ret = await hub.exec.boto3.client.iam.delete_policy(
                ctx, PolicyArn=resource_id if resource_id else name
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted aws.iam.policy '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists all of your own customer-defined managed policies.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.policy
    """

    result = {}
    # Set scope to local to only list the customer-defined managed policies.
    ret_policies = await hub.exec.boto3.client.iam.list_policies(ctx, Scope="Local")

    if not ret_policies["result"]:
        hub.log.debug(f"Could not describe policy {ret_policies['comment']}")
        return {}

    for policy in ret_policies["ret"]["Policies"]:
        translated_resource = (
            await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
                ctx, policy
            )
        )
        if translated_resource is None:
            continue

        result[translated_resource["resource_id"]] = {
            "aws.iam.policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



"""

Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_security_group
hub.exec.boto3.client.ec2.delete_security_group
hub.exec.boto3.client.ec2.describe_security_groups
hub.tool.boto3.resource.exec(resource, authorize_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, authorize_ingress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_ingress, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.ec2.security_group.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    group_id: str,
    ip_protocol: str,
    from_port: int,
    to_port: int,
    is_egress: bool,
    resource_id: str = None,
    cidr_ipv4: str = None,
    cidr_ipv6: str = None,
    prefix_list_id: str = None,
    referenced_group_info: Dict = None,
    tags: list = None,
    description: str = None,
    cidr_ipv4_desc: str = None,
    cidr_ipv6_desc: str = None,
    prefix_list_id_desc: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a security group rule.
    authorize_security_group_ingress: Adds the specified inbound (ingress) rules to a security group. An inbound rule
    permits instances to receive traffic from the specified IPv4 or IPv6 CIDR address range, or from the instances that
    are associated with the specified destination security groups.You specify a protocol for each rule (for example, TCP).
    For TCP and UDP, you must also specify the destination port or port range. For ICMP/ICMPv6, you must also specify the
    ICMP/ICMPv6 type and code. You can use -1 to mean all types or all codes. Rule changes are propagated to instances within
    the security group as quickly as possible. However, a small delay might occur.
    authorize_security_group_egress: [VPC only] Adds the specified outbound (egress) rules to a security group for use with a VPC.
    An outbound rule permits instances to send traffic to the specified IPv4 or IPv6 CIDR address ranges, or to the instances that
    are associated with the specified source security groups. You specify a protocol for each rule (for example, TCP). For the TCP
    and UDP protocols, you must also specify the destination port or port range. For the ICMP protocol, you must also specify the
    ICMP type and code. You can use -1 for the type or code to mean all types or all codes. Rule changes are propagated to affected
    instances as quickly as possible. However, a small delay might occur.

    Args:
        hub:
        ctx:
        name(Text): An Idem name of the resource.
        group_id(str): The ID of the security group
        is_egress(bool): To find the type of rule, whether it is a ingress or egress.
        ip_protocol(str): The IP protocol name (tcp , udp , icmp , icmpv6 ) or number (see Protocol Numbers ). [VPC only] Use -1 to specify all protocols.
                          When authorizing security group rules, specifying -1 or a protocol number other than tcp , udp , icmp , or icmpv6 allows traffic
                          on all ports, regardless of any port range you specify. For tcp , udp , and icmp , you must specify a port range. For icmpv6 ,
                          the port range is optional; if you omit the port range, traffic for all types and codes is allowed.
        from_port(int): The start of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 type number. A value of -1 indicates all ICMP/ICMPv6 types.
                        If you specify all ICMP/ICMPv6 types, you must specify all codes.
        to_port(int): The end of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 code. A value of -1 indicates all ICMP/ICMPv6 codes.
                      If you specify all ICMP/ICMPv6 types, you must specify all codes.
        cidr_ipv4(str, optional): The IPv4 CIDR range. You can either specify a CIDR range or a source security group, not both.
                        To specify a single IPv4 address, use the /32 prefix length.
        resource_id(Text, optional): AWS Security group rule ID
        cidr_ipv6(str, optional): The IPv6 CIDR range. You can either specify a CIDR range or a source security group, not both.
                                  To specify a single IPv6 address, use the /128 prefix length.
        prefix_list_id(str, optional): The ID of the prefix.
        referenced_group_info(Dict, optional): The security group and Amazon Web Services account ID pairs.
        description(str, optional): The description of the security group rule.
        cidr_ipv4_desc(str, optional): A description for the security group rule that references this IPv4 address range.
                                       Constraints: Up to 255 characters in length. Allowed characters are a-z, A-Z, 0-9, spaces, and ._-:/()#,@[]+=&;{}!$*
        cidr_ipv6_desc(str, optional): A description for the security group rule that references this IPv6 address range.
                                    Constraints: Up to 255 characters in length. Allowed characters are a-z, A-Z, 0-9, spaces, and ._-:/()#,@[]+=&;{}!$*
        prefix_list_id_desc(str, optional): A description for the security group rule that references this prefix list ID.
                                        Constraints: Up to 255 characters in length. Allowed characters are a-z, A-Z, 0-9, spaces, and ._-:/()#,@[]+=;{}!$*
        tags(Dict, Optional): List of Name, Value pair to tag security group

    Request Syntax:
        [security_group_rule_name]:
          aws.ec2.security_group_rule.present:
            - group_id: 'string'
            - is_egress: 'bool'
            - ip_protocol: 'string'
            - from_port: 'int'
            - to_port: 'int'
            - cidr_ipv4: 'string'
            - cidr_ipv6: 'string'
            - prefix_list_id: 'string'
            - referenced_group_info: 'dict'
            - tags:
                - Key: 'string'
                  Value: 'string'
            - description: 'string'
            - cidr_ipv4_desc: 'string'
            - cidr_ipv6_desc: 'string'
            - prefix_list_id_desc: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-sg-rule:
              aws.ec2.security_group_rule.present:
                - group_id: sg-0dd442ba9f89c5d59
                - is_egress: false
                - ip_protocol: tcp
                - from_port: 60
                - to_port: 60
                - cidr_ipv4: 0.0.0.0/0
                - tags:
                    - Key: test_name5
                      Value: test-rule5
                - description: 'Security group desc'
                - cidr_ipv4_desc: 'This is new cidr_ipv4_desc'
                - cidr_ipv6_desc: 'This is new cidr_ipv6_desc'
                - prefix_list_id_desc: 'This is new prefix_list_id_desc'
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.ec2.describe_security_group_rules(
            ctx, SecurityGroupRuleIds=[resource_id]
        )
    if before and before["result"] and before["ret"]["SecurityGroupRules"]:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_rule_to_present(
            raw_resource=before["ret"]["SecurityGroupRules"][0], idem_resource_name=name
        )
        plan_state = copy.deepcopy(result["old_state"])
        result["comment"] = f"'{name}' already present."
        security_group_rules = []
        updated_rule = {}
        new_modified_rule = dict(SecurityGroupRuleId=None, SecurityGroupRule=None)
        new_modified_rule["SecurityGroupRuleId"] = name
        if ip_protocol:
            updated_rule["IpProtocol"] = ip_protocol
        if from_port:
            updated_rule["FromPort"] = from_port
        if to_port:
            updated_rule["ToPort"] = to_port
        if cidr_ipv4:
            updated_rule["CidrIpv4"] = cidr_ipv4
        if cidr_ipv6:
            updated_rule["CidrIpv6"] = cidr_ipv6
        if prefix_list_id:
            updated_rule["PrefixListId"] = prefix_list_id
        if referenced_group_info:
            updated_rule["ReferencedGroupId"] = referenced_group_info
        if description:
            updated_rule["Description"] = description
        new_modified_rule["SecurityGroupRule"] = updated_rule
        security_group_rules.append(new_modified_rule)
        if ctx.get("test", False):
            if security_group_rules is not None:
                plan_state["security_group_rules"] = security_group_rules
            if group_id is not None:
                plan_state["group_id"] = group_id
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.ec2.modify_security_group_rules(
                ctx,
                GroupId=group_id,
                SecurityGroupRules=security_group_rules,
            )
            result["comment"] = update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "group_id": group_id,
                    "ip_protocol": ip_protocol,
                    "from_port": from_port,
                    "to_port": to_port,
                    "is_egress": is_egress,
                    "cidr_ipv4": cidr_ipv4,
                    "tags": tags,
                    "description": description,
                    "cidr_ipv4_desc": cidr_ipv4_desc,
                    "cidr_ipv6_desc": cidr_ipv6_desc,
                    "prefix_list_id_desc": prefix_list_id_desc,
                },
            )
            result["comment"] = f"Would create aws.ec2.security_group_rule {name}"
            return result
        try:
            ip_permissions = []
            ip_perm = {}
            if ip_protocol:
                ip_perm["IpProtocol"] = ip_protocol
            if from_port:
                ip_perm["FromPort"] = from_port
            if to_port:
                ip_perm["ToPort"] = to_port
            if cidr_ipv4:
                IpRanges = {}
                if cidr_ipv4_desc:
                    IpRanges["Description"] = cidr_ipv4_desc
                if cidr_ipv4:
                    IpRanges["CidrIp"] = cidr_ipv4
                ip_perm["IpRanges"] = [IpRanges]
            if cidr_ipv6:
                Ipv6Ranges = {}
                if cidr_ipv6_desc:
                    Ipv6Ranges["Description"] = cidr_ipv6_desc
                if cidr_ipv6:
                    Ipv6Ranges["CidrIpv6"] = cidr_ipv6
                ip_perm["Ipv6Ranges"] = [Ipv6Ranges]
            if prefix_list_id:
                prefix_list = {}
                if prefix_list_id_desc:
                    prefix_list["Description"] = prefix_list_id_desc
                if prefix_list_id:
                    prefix_list["PrefixListId"] = prefix_list_id
                ip_perm["PrefixListIds"] = [prefix_list]
            if referenced_group_info:
                ip_perm["UserIdGroupPairs"] = [referenced_group_info]
            ip_permissions.append(ip_perm)
            if not is_egress:
                ret = await hub.exec.boto3.client.ec2.authorize_security_group_ingress(
                    ctx,
                    GroupId=group_id,
                    IpPermissions=ip_permissions,
                    TagSpecifications=[
                        {"ResourceType": "security-group-rule", "Tags": tags}
                    ]
                    if tags is not None
                    else None,
                )
                result["result"] = ret["result"]
                if not ret["result"]:
                    result["comment"] = ret["comment"]
                    return result
                result["comment"] = f"Created '{name}'"
            else:
                ret = await hub.exec.boto3.client.ec2.authorize_security_group_egress(
                    ctx,
                    GroupId=group_id,
                    IpPermissions=ip_permissions,
                    TagSpecifications=[
                        {"ResourceType": "security-group-rule", "Tags": tags}
                    ]
                    if tags is not None
                    else None,
                )
                result["result"] = ret["result"]
                if not ret["result"]:
                    result["comment"] = ret["comment"]
                    return result
                result["comment"] = f"Created '{name}'"
            resource_id = ret["ret"]["SecurityGroupRules"][0]["SecurityGroupRuleId"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.ec2.describe_security_group_rules(
                ctx,
                SecurityGroupRuleIds=[resource_id],
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_rule_to_present(
                raw_resource=after["ret"]["SecurityGroupRules"][0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    group_id: str,
    resource_id: str,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a security group rule.
    revoke_ingress: Removes the specified inbound (ingress) rules from a security group. You can specify rules using either
    rule IDs or security group rule properties. If you use rule properties, the values that you specify (for example, ports)
    must match the existing rule's values exactly. Each rule has a protocol, from and to ports, and source (CIDR range,
    security group, or prefix list). For the TCP and UDP protocols, you must also specify the destination port or range of ports.
    For the ICMP protocol, you must also specify the ICMP type and code. If the security group rule has a description, you do
    not need to specify the description to revoke the rule. [EC2-Classic, default VPC] If the values you specify do not match
    the existing rule's values, no error is returned, and the output describes the security group rules that were not revoked.
    Amazon Web Services recommends that you describe the security group to verify that the rules were removed.
    Rule changes are propagated to instances within the security group as quickly as possible. However, a small delay might occur.
    revoke_egress: [VPC only] Removes the specified outbound (egress) rules from a security group for EC2-VPC. This action does not
    apply to security groups for use in EC2-Classic. You can specify rules using either rule IDs or security group rule properties.
    If you use rule properties, the values that you specify (for example, ports) must match the existing rule's values exactly.
    Each rule has a protocol, from and to ports, and destination (CIDR range, security group, or prefix list). For the TCP and UDP
    protocols, you must also specify the destination port or range of ports. For the ICMP protocol, you must also specify the ICMP
    type and code. If the security group rule has a description, you do not need to specify the description to revoke the rule.
    [Default VPC] If the values you specify do not match the existing rule's values, no error is returned, and the output describes
    the security group rules that were not revoked. Amazon Web Services recommends that you describe the security group to verify
    that the rules were removed. Rule changes are propagated to instances within the security group as quickly as possible. However,
    a small delay might occur.

    Args:
        hub:
        ctx:
        name(Text): A name, ID to identify the resource.
        group_id(str): The ID of the Security group
        resource_id(Text): The AWS ID of the Security group rule

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.security_group_rule.absent:
                - name: value
                - group_id: "sg-0008bd25b7867b5cf"
    """
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "SecurityGroup", group_id)
    before = await hub.exec.boto3.client.ec2.describe_security_group_rules(
        ctx, SecurityGroupRuleIds=[resource_id]
    )
    if not before and not before["result"]:
        result["comment"] = f"'{name}' is already in deleted state."
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_rule_to_present(
            raw_resource=before["ret"]["SecurityGroupRules"][0], idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = f"Would delete aws.ec2.security_group_rule {name}"
            return result
        try:
            # checks for egress or ingress using before's output and revoke function needed sg resource
            if not before["ret"]["SecurityGroupRules"][0]["IsEgress"]:
                ret = resource.revoke_ingress(SecurityGroupRuleIds=[resource_id])
                result["result"] = ret["Return"]
                if not result["result"]:
                    result["comment"] = f"'{name}' is already in deleted state."
                    return result
                result["comment"] = f"Deleted '{name}'"
            else:
                ret = resource.revoke_egress(SecurityGroupRuleIds=[resource_id])
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = f"'{name}' is already in deleted state."
                    return result
                result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = f"{e.__class__.__name__}: {e}"

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Describes one or more of security group rules. Allows to use filter based on resource type i.e.
    group-id, security-group-rule-id and tags.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.security_group_rule
    """
    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_security_group_rules(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe security_group rule {ret['comment']}")
        result["comment"] = ret["comment"]
        result["result"] = False
        return result

    for resource in ret["ret"]["SecurityGroupRules"]:
        resource_id = resource.get("SecurityGroupRuleId")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_sg_rule_to_present(
                raw_resource=resource, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.security_group_rule.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.organizations.deregister_delegated_administrator
hub.exec.boto3.client.organizations.list_delegated_administrators
hub.exec.boto3.client.organizations.register_delegated_administrator
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.organizations.account.absent",
        ],
    },
    "present": {
        "require": [
            "aws.organizations.organization.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    parent_id: str,
    tags: List = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an organizational unit (OU) within a root or parent OU. An OU is a container
    for accounts that enables you to organize your accounts to apply policies according
    to your business requirements. The number of levels deep that you can nest OUs is
    dependent upon the policy types enabled for that root. For service control policies,
    the limit is five.

    Args:
        name(Text): Name of the resource.
        parent_id(Text): The unique identifier (ID) of the parent root or OU that you
                                  want to create the new OU in.
        tags(list,Optional): A list of tags that you want to attach to the newly created OU
        resource_id(Text, Optional): AWS Organization Unit ID to identify the resource


    Request Syntax:
        [org-id]:
          aws.organizations.organization_unit.present:
          - parent_id: 'string'
          - name: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            ou-examplerootid111-exampleouid111:
              aws.organizations.organization_unit.present:
                - parent_id: ou-01919
                - name: ou-new
                - tags:
                    - Key: org
                      Value: guardrails
                - resource_id: ou-examplerootid111-exampleouid111
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    existing_tags = None

    if resource_id:
        before = await hub.exec.boto3.client.organizations.describe_organizational_unit(
            ctx, OrganizationalUnitId=resource_id
        )
    update = False

    if not before:
        # organization unit does not exist, create

        try:

            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "parent_id": parent_id,
                        "name": name,
                        "tags": tags,
                    },
                )
                result["comment"] = result["comment"] + (
                    f"Would create aws.organizations.organization_unit '{name}'.",
                )
                return result

            create_ret = (
                await hub.exec.boto3.client.organizations.create_organizational_unit(
                    ctx, ParentId=parent_id, Name=name, Tags=tags
                )
            )

            result["result"] = create_ret["result"]

            if not result["result"]:
                result["comment"] = result["comment"] + create_ret["comment"]
                return result
            resource_id = create_ret["ret"]["OrganizationalUnit"]["Id"]
            result["comment"] = result["comment"] + (
                f"Created aws.organizations.organization_unit '{name}'.",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        # organizational_unit exists , update
        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would update aws.organizations.organization_unit '{name}'.",
            )
        existing_ou_name = before["ret"]["OrganizationalUnit"]["Name"]

        old_tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=resource_id
        )
        if not old_tags["result"]:
            hub.log.debug(
                f"Unable to list tags for resource {resource_id} with error: {old_tags['comment']}"
            )
            result["comment"] = result["comment"] + old_tags["comment"]
            result["result"] = old_tags["result"]
        else:
            before["ret"]["OrganizationalUnit"]["Tags"] = (
                old_tags["ret"]["Tags"] if old_tags else None
            )
            existing_tags = old_tags["ret"]["Tags"]

        result[
            "old_state"
        ] = hub.tool.aws.organization.conversion_utils.convert_raw_ou_to_present(
            before["ret"]["OrganizationalUnit"],
            parent_id,
        )
        plan_state = copy.deepcopy(result["old_state"])

        try:

            if existing_ou_name != name:

                if ctx.get("test", False):
                    plan_state["name"] = name
                else:
                    update_ret = await hub.exec.boto3.client.organizations.update_organizational_unit(
                        ctx,
                        OrganizationalUnitId=resource_id,
                        Name=name,
                    )

                    result["result"] = update_ret["result"]
                    if not result["result"]:
                        result["comment"] = result["comment"] + update_ret["comment"]
                        return result
                    result["comment"] = result["comment"] + (
                        f"Updated ou name on aws.organizations.organization_unit '{name}'.",
                    )

                update = True

            if tags is not None:
                update_tags_ret = (
                    await hub.exec.aws.organizations.organization.update_tags(
                        ctx, resource_id, existing_tags if existing_tags else None, tags
                    )
                )

                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]

                    result["result"] = False
                    return result

                elif update_tags_ret["ret"] is not None:
                    update = True
                    result["comment"] = result["comment"] + (
                        "Updated tags on aws.organizations.organization_unit '{name}'.",
                    )

                if ctx.get("test", False) and update_tags_ret["ret"]:
                    plan_state["tags"] = update_tags_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update aws.organizations.organization_unit '{name}'",
                    )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:

        if ctx.get("test", False):
            result["new_state"] = plan_state
            return result

        if not before or update:

            after = (
                await hub.exec.boto3.client.organizations.describe_organizational_unit(
                    ctx, OrganizationalUnitId=resource_id
                )
            )

            if after.get("ret"):
                updated_tag = (
                    await hub.exec.boto3.client.organizations.list_tags_for_resource(
                        ctx, ResourceId=resource_id
                    )
                )
                if not updated_tag["result"]:
                    hub.log.debug(
                        f"Unable to list tags for resource {resource_id} with error: {updated_tag['comment']}"
                    )
                    result["comment"] = result["comment"] + updated_tag["comment"]
                    result["result"] = updated_tag["result"]
                after["ret"]["OrganizationalUnit"]["Tags"] = (
                    updated_tag["ret"].get("Tags", [])
                    if updated_tag and updated_tag.get("ret")
                    else None
                )
                result[
                    "new_state"
                ] = hub.tool.aws.organization.conversion_utils.convert_raw_ou_to_present(
                    after["ret"]["OrganizationalUnit"],
                    parent_id,
                )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False

    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes an organizational unit (OU) from a root or another OU. You must first remove all accounts
    and child OUs from the OU that you want to delete.This operation can be called only from the
    organization's management account.

    Args:
        name(Text): Name of the resource.
        resource_id(Text): AWS Organization Unit ID to identify the resource

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.organizations.organization_unit.absent:
                - name: ou-id1011
                - resource_id: ou-examplerootid111-exampleouid111
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_organizational_unit(
        ctx, OrganizationalUnitId=resource_id
    )

    if not before:
        result["comment"] = result["comment"] + (
            f"aws.organizations.organization_unit '{name}' already absent",
        )
    else:

        result[
            "old_state"
        ] = hub.tool.aws.organization.conversion_utils.convert_raw_ou_to_present(
            before["ret"]["OrganizationalUnit"]
        )

        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would delete aws.organizations.organization_unit '{name}'",
            )
            return result
        try:
            ret = await hub.exec.boto3.client.organizations.delete_organizational_unit(
                ctx, OrganizationalUnitId=resource_id
            )

            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = False
                return result
            result["comment"] = result["comment"] + (
                f"aws.organizations.organization_unit '{name}' deleted.",
            )
            result["result"] = ret["result"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Lists the organizational units (OUs) in a parent organizational unit or root.

    Idem does not support passing arguments to the describe function as of today , hence describe organizational unit
    will describe all the organizational unit one level under the root of the organization.
    In future if Idem starts supporting input arguments , we can pass any parent_id and this function will list
    all organizational units one level under the parent.


    Returns:
        Dict[str, Any]


    Examples:

        .. code-block:: bash

            $ idem describe aws.organizations.organization_unit
    """

    result = {}

    list_roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

    if not (list_roots_resp and list_roots_resp["ret"].get("Roots")):
        return result

    parent_id = list_roots_resp["ret"]["Roots"][0]["Id"]

    if parent_id is None:
        return result

    org_units = (
        await hub.exec.boto3.client.organizations.list_organizational_units_for_parent(
            ctx, ParentId=parent_id
        )
    )

    if not org_units["result"]:
        hub.log.debug(
            f"Could not describe organization units for parent_id {parent_id} with error: {org_units['comment']}"
        )
        return {}

    organizational_units = org_units["ret"]["OrganizationalUnits"]

    for ou in organizational_units:
        tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=ou["Id"]
        )
        if not tags["result"]:
            hub.log.debug(
                f"Unable to list tags for resource {ou['Id']} with error: {tags['comment']}"
            )
            continue
        ou["Tags"] = tags["ret"].get("Tags") if tags else None
        translated_resource = (
            hub.tool.aws.organization.conversion_utils.convert_raw_ou_to_present(
                ou, parent_id
            )
        )

        result[ou["Name"]] = {
            "aws.organizations.organization_unit.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_cidr_blocks(
    hub,
    ctx,
    vpc_id: str,
    old_ipv4_cidr_blocks: List[Dict[str, Any]],
    old_ipv6_cidr_blocks: List[Dict[str, Any]],
    new_ipv4_cidr_blocks: List[Dict[str, Any]],
    new_ipv6_cidr_blocks: List[Dict[str, Any]],
):
    """
    Update associated cidr blocks of a vpc. This function compares the existing(old) cidr blocks and the
    new cidr blocks. Cidr blocks that are in the new_cidr_blocks but not in the old_cidr_blocks will be associated to
    vpc. Cidr blocks that are in the old_cidr_blocks but not in the new_cidr_blocks will be disassociated from vpc.

    Args:
        hub:
        ctx:
        vpc_id: The AWS resource id of the existing vpc
        old_ipv4_cidr_blocks: The ipv4 cidr blocks on the existing vpc
        old_ipv6_cidr_blocks: The ipv6 cidr blocks on the existing vpc
        new_ipv4_cidr_blocks: The expected ipv4 cidr blocks on the existing vpc
        new_ipv6_cidr_blocks: The expected ipv6 cidr blocks on the existing vpc

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    # If a block is None, we'll skip updating such cidr block.
    if new_ipv4_cidr_blocks is None:
        new_ipv4_cidr_blocks = old_ipv4_cidr_blocks
    if new_ipv6_cidr_blocks is None:
        new_ipv6_cidr_blocks = old_ipv6_cidr_blocks
    result = dict(comment=(), result=True, ret=None)
    ipv4_cidr_block_map = {
        cidr_block.get("CidrBlock"): cidr_block for cidr_block in old_ipv4_cidr_blocks
    }
    ipv6_cidr_block_map = {
        cidr_block.get("Ipv6CidrBlock"): cidr_block
        for cidr_block in old_ipv6_cidr_blocks
    }
    cidr_blocks_to_add = []
    ipv4_cidr_block_results = copy.deepcopy(ipv4_cidr_block_map)
    ipv6_cidr_block_results = copy.deepcopy(ipv6_cidr_block_map)
    for cidr_block_set in new_ipv4_cidr_blocks:
        if cidr_block_set.get("CidrBlock") not in ipv4_cidr_block_map:
            cidr_blocks_to_add.append(
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    cidr_block_set, "ipv4"
                )
            )
            ipv4_cidr_block_results[cidr_block_set.get("CidrBlock")] = cidr_block_set
        else:
            ipv4_cidr_block_map.pop(cidr_block_set.get("CidrBlock"), None)
    for ipv6_cidr_block_set in new_ipv6_cidr_blocks:
        if ipv6_cidr_block_set.get("Ipv6CidrBlock") not in ipv6_cidr_block_map:
            cidr_blocks_to_add.append(
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    ipv6_cidr_block_set, "ipv6"
                )
            )
            ipv6_cidr_block_results[
                ipv6_cidr_block_set.get("Ipv6CidrBlock")
            ] = ipv6_cidr_block_set
        else:
            ipv6_cidr_block_map.pop(ipv6_cidr_block_set.get("Ipv6CidrBlock"), None)
            ipv6_cidr_block_results.pop(ipv6_cidr_block_set.get("Ipv6CidrBlock"), None)
    cidr_blocks_to_remove = list(ipv4_cidr_block_map.values()) + list(
        ipv6_cidr_block_map.values()
    )
    if (not cidr_blocks_to_remove) and (not cidr_blocks_to_add):
        return result
    if not ctx.get("test", False):
        for cidr_block in cidr_blocks_to_remove:
            ret = await hub.exec.boto3.client.ec2.disassociate_vpc_cidr_block(
                ctx, AssociationId=cidr_block.get("AssociationId")
            )
            if not ret.get("result"):
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
        for request_payload in cidr_blocks_to_add:
            ret = await hub.exec.boto3.client.ec2.associate_vpc_cidr_block(
                ctx, VpcId=vpc_id, **request_payload
            )
            if not ret.get("result"):
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
    result["comment"] = (
        f"Update tags: Add [{cidr_blocks_to_add}] Remove [{cidr_blocks_to_remove}]",
    )
    result["ret"] = {
        "cidr_block_association_set": list(ipv4_cidr_block_results.values()),
        "ipv6_cidr_block_association_set": list(ipv6_cidr_block_results.values()),
    }
    return result


async def update_vpc_attributes(
    hub,
    ctx,
    enable_dns_hostnames,
    enable_dns_support,
    resource_id,
):
    result = dict(comment=(), result=True, ret=None)
    update_ret = await existing_attributes(hub, ctx, resource_id)
    if not update_ret["result"]:
        result["comment"] = update_ret["comment"]
        result["result"] = False
    if (
        enable_dns_support is not None
        and update_ret["result"]
        and update_ret["ret"].get("existingEnableDnsSupport") is not None
        and enable_dns_support != update_ret["ret"].get("existingEnableDnsSupport")
    ):
        if not ctx.get("test", False):
            modify_enable_dns_support = (
                await hub.exec.boto3.client.ec2.modify_vpc_attribute(
                    ctx,
                    EnableDnsSupport={"Value": enable_dns_support},
                    VpcId=resource_id,
                )
            )
            result["comment"] = (
                f"Updated DNS support attribute to {enable_dns_support} on vpc {resource_id}",
            )
            result["result"] = result["result"] and modify_enable_dns_support["result"]
    if (
        enable_dns_hostnames is not None
        and update_ret["result"]
        and update_ret["ret"].get("existingEnableDnsHostnames") is not None
        and enable_dns_hostnames != update_ret["ret"].get("existingEnableDnsHostnames")
        and enable_dns_support
    ):
        if not ctx.get("test", False):
            modify_enable_dns_hostnames = (
                await hub.exec.boto3.client.ec2.modify_vpc_attribute(
                    ctx,
                    EnableDnsHostnames={"Value": enable_dns_hostnames},
                    VpcId=resource_id,
                )
            )
            result["comment"] = (
                f"Updated DNS hostname attribute to {enable_dns_hostnames} on vpc {resource_id}",
            )
            result["result"] = (
                result["result"] and modify_enable_dns_hostnames["result"]
            )

    update_ret = await existing_attributes(hub, ctx, resource_id)
    if not update_ret["result"]:
        result["comment"] = update_ret["comment"]
        result["result"] = False
    else:
        result["ret"] = {
            "enable_dns_support": update_ret["ret"].get("existingEnableDnsSupport"),
            "enable_dns_hostnames": update_ret["ret"].get("existingEnableDnsHostnames"),
        }
    return result


async def existing_attributes(hub, ctx, resource_id):
    result = dict(comment=(), result=True, ret=None)
    dnsHostName = None
    dnsSupport = None
    if not ctx.get("test", False):
        existingEnableDnsHostnames = (
            await hub.exec.boto3.client.ec2.describe_vpc_attribute(
                ctx,
                Attribute="enableDnsHostnames",
                VpcId=resource_id,
            )
        )
        if existingEnableDnsHostnames["result"]:
            dnsHostName = existingEnableDnsHostnames["ret"]["EnableDnsHostnames"][
                "Value"
            ]
        else:
            result["comment"] = existingEnableDnsHostnames["comment"]
            result["result"] = False

    if not ctx.get("test", False):
        existingEnableDnsSupport = (
            await hub.exec.boto3.client.ec2.describe_vpc_attribute(
                ctx,
                Attribute="enableDnsSupport",
                VpcId=resource_id,
            )
        )
        if existingEnableDnsSupport["result"]:
            dnsSupport = existingEnableDnsSupport["ret"]["EnableDnsSupport"]["Value"]
        else:
            result["comment"] = existingEnableDnsSupport["comment"]
            result["result"] = False
    result["ret"] = {
        "existingEnableDnsHostnames": dnsHostName,
        "existingEnableDnsSupport": dnsSupport,
    }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS EC2 to present input format.
"""


async def convert_raw_vpc_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("VpcId")
    resource_parameters = OrderedDict(
        {
            "InstanceTenancy": "instance_tenancy",
            "Tags": "tags",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    # The following code block is to make sure to only output the associated/associating cidr blocks and
    # dis-regard the disassociated cidr blocks.
    if raw_resource.get("CidrBlockAssociationSet"):
        ipv4_cidr_block_association_set = []
        for cidr_block in raw_resource.get("CidrBlockAssociationSet"):
            if cidr_block.get("CidrBlockState").get("State") in [
                "associated",
                "associating",
            ]:
                ipv4_cidr_block_association_set.append(cidr_block)
        resource_translated[
            "cidr_block_association_set"
        ] = ipv4_cidr_block_association_set
    if resource_id:
        enableDnsHostnames = await hub.exec.boto3.client.ec2.describe_vpc_attribute(
            ctx, Attribute="enableDnsHostnames", VpcId=resource_id
        )
        if enableDnsHostnames and enableDnsHostnames["result"] is True:
            resource_translated["enable_dns_hostnames"] = enableDnsHostnames["ret"][
                "EnableDnsHostnames"
            ]["Value"]
        else:
            # TODO - Need to handle the error efficiently.
            hub.log.warning(
                f"Failed on fetching enableDnsHostnames on vpc {resource_id} with error {enableDnsHostnames['comment']}."
            )
        enableDnsSupport = await hub.exec.boto3.client.ec2.describe_vpc_attribute(
            ctx, Attribute="enableDnsSupport", VpcId=resource_id
        )
        if enableDnsSupport and enableDnsSupport["result"] is True:
            resource_translated["enable_dns_support"] = enableDnsSupport["ret"][
                "EnableDnsSupport"
            ]["Value"]
        else:
            # TODO - Need to handle the error efficiently.
            hub.log.warning(
                f"Failed on fetching enableDnsSupport on vpc {resource_id} with error {enableDnsSupport['comment']}."
            )
    if raw_resource.get("Ipv6CidrBlockAssociationSet"):
        ipv6_cidr_block_association_set = []
        for cidr_block in raw_resource.get("Ipv6CidrBlockAssociationSet"):
            if cidr_block.get("Ipv6CidrBlockState").get("State") in [
                "associated",
                "associating",
            ]:
                if "NetworkBorderGroup" in cidr_block:
                    # Translate describe output to the correct present input format
                    ipv6_cidr_block_network_border_group = cidr_block.pop(
                        "NetworkBorderGroup"
                    )
                    cidr_block[
                        "Ipv6CidrBlockNetworkBorderGroup"
                    ] = ipv6_cidr_block_network_border_group
                ipv6_cidr_block_association_set.append(cidr_block)
        resource_translated[
            "ipv6_cidr_block_association_set"
        ] = ipv6_cidr_block_association_set
    return resource_translated


def convert_raw_subnet_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("SubnetId")
    resource_parameters = OrderedDict(
        {
            "VpcId": "vpc_id",
            "CidrBlock": "cidr_block",
            "AvailabilityZone": "availability_zone",
            "OutpostArn": "outpost_arn",
            "Tags": "tags",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if (not raw_resource.get("AvailabilityZone")) and raw_resource.get(
        "AvailabilityZoneId"
    ):
        # Only populate availability_zone_id field when availability_zone doesn't exist
        resource_translated["availability_zone_id"] = raw_resource.get(
            "AvailabilityZoneId"
        )
    if raw_resource.get("Ipv6CidrBlockAssociationSet"):
        ipv6_cidr_block_association_set = (
            hub.tool.aws.network_utils.get_associated_ipv6_cidr_blocks(
                raw_resource.get("Ipv6CidrBlockAssociationSet")
            )
        )
        # We should only output the associated ipv6 cidr block, and theoretically there should only be one,
        # since AWS only supports one ipv6 cidr block association on a subnet
        if ipv6_cidr_block_association_set:
            resource_translated["ipv6_cidr_block"] = ipv6_cidr_block_association_set[
                0
            ].get("Ipv6CidrBlock")
    return resource_translated


def convert_raw_sg_rule_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("SecurityGroupRuleId")
    resource_parameters = OrderedDict(
        {
            "GroupId": "group_id",
            "IsEgress": "is_egress",
            "IpProtocol": "ip_protocol",
            "FromPort": "from_port",
            "ToPort": "to_port",
            "CidrIpv4": "cidr_ipv4",
            "CidrIpv6": "cidr_ipv6",
            "Tags": "tags",
            "PrefixListId": "prefix_list_id",
            "Description": "description",
            "ReferencedGroupInfo": "referenced_group_info",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated


def convert_raw_transit_gateway_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("TransitGatewayId")
    resource_parameters = OrderedDict(
        {
            "Description": "description",
            "Options": "options",
            "Tags": "tags",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


def convert_raw_flow_log_to_present(hub, flow_log: Dict):
    """
    Convert the aws flow log response to a common format

    Args:
        hub: required for functions in hub
        flow_log: describe flow log response

    Returns:
        A dictionary of flow log
    """
    describe_parameters = OrderedDict(
        {
            "DeliverLogsPermissionArn": "iam_role",
            "LogGroupName": "log_group_name",
            "TrafficType": "traffic_type",
            "LogDestinationType": "log_destination_type",
            "LogDestination": "log_destination",
            "LogFormat": "log_format",
            "MaxAggregationInterval": "max_aggregation_interval",
            "Tags": "tags",
        }
    )
    translated_flow_log = {}
    flow_log_id = flow_log.get("FlowLogId")
    resource_id = flow_log.get("ResourceId")
    resource_ids = [resource_id]
    resource_type = resource_id.split("-")[0]
    translated_flow_log["resource_ids"] = resource_ids
    if resource_type == "vpc":
        resource_type = "VPC"
    elif resource_type == "subnet":
        resource_type = "Subnet"
    else:
        resource_type = "NetworkInterface"
    translated_flow_log["resource_type"] = resource_type
    translated_flow_log["resource_id"] = flow_log_id

    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if flow_log.get(parameter_old_key) is not None:
            translated_flow_log[parameter_new_key] = flow_log.get(parameter_old_key)

    return translated_flow_log


def convert_raw_route_table_to_present(
    hub, resource: Dict, idem_resource_name: str = None
) -> Dict[str, Any]:
    describe_parameters = OrderedDict(
        {
            "VpcId": "vpc_id",
            "PropagatingVgws": "propagating_vgws",
            "Tags": "tags",
        }
    )
    if not resource:
        return {}
    resource_id = resource.get("RouteTableId")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if resource.get(parameter_old_key) is not None:
            resource_translated[parameter_new_key] = resource.get(parameter_old_key)

    # Add only required parameters in Describe
    routes_to_add = []
    if resource.get("Routes"):
        for route_to_add in resource.get("Routes"):
            if route_to_add.get("State") == "active":
                del route_to_add["Origin"]
                routes_to_add.append(route_to_add)
        resource_translated["routes"] = routes_to_add

    associations_to_add = []
    if resource.get("Associations"):
        for association_to_add in resource.get("Associations"):
            if (
                association_to_add.get("AssociationState").get("State") == "associated"
                and not association_to_add["Main"]
            ):
                associations_to_add.append(association_to_add)
        resource_translated["associations"] = associations_to_add
    return resource_translated


def convert_raw_tg_vpc_attachment(
    hub, raw_resource: Dict, resource_name: str = None
) -> Dict[str, any]:
    resource_id = raw_resource.get("TransitGatewayAttachmentId")
    translated_resource = {"name": resource_name, "resource_id": resource_id}

    resource_parameters_transit_gateway = OrderedDict(
        {
            "TransitGatewayId": "transit_gateway",
            "VpcId": "vpc_id",
            "SubnetIds": "subnet_ids",
            "Tags": "tags",
            "Options": "options",
        }
    )

    for camel_case_key, snake_case_key in resource_parameters_transit_gateway.items():
        if raw_resource.get(camel_case_key):
            translated_resource[snake_case_key] = raw_resource.get(camel_case_key)
    return translated_resource


def convert_raw_dhcp_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("DhcpOptionsId")
    resource_parameters = OrderedDict({"Tags": "tags"})
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if raw_resource.get("DhcpConfigurations"):
        dhcp_configs = []
        for dhcp_conf in raw_resource.get("DhcpConfigurations"):
            dhcp_config_values_list = []
            for dhcp in dhcp_conf.get("Values"):
                if "Value" in dhcp:
                    dhcp_config_values_list.append(dhcp.get("Value"))
            dhcp_config = {
                "Key": dhcp_conf.get("Key"),
                "Values": dhcp_config_values_list,
            }
            dhcp_configs.append(dhcp_config)
        resource_translated["dhcp_configurations"] = dhcp_configs
    return resource_translated


def convert_raw_sg_to_present(hub, security_group: Dict):
    """
    Convert the aws security group response to a common format

    Args:
        hub: required for functions in hub
        security_group: describe security group response

    Returns:
        A dictionary of sg group
    """
    translated_security_group = {}
    describe_parameters = OrderedDict(
        {
            "GroupId": "resource_id",
            "GroupName": "name",
            "VpcId": "vpc_id",
            "Tags": "tags",
            "Description": "description",
        }
    )
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if security_group.get(parameter_old_key) is not None:
            translated_security_group[parameter_new_key] = security_group.get(
                parameter_old_key
            )

    return translated_security_group


def convert_raw_internet_gateway_to_present(
    hub, resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource.get("InternetGatewayId"),
    }
    if resource.get("Tags"):
        resource_translated["tags"] = resource.get("Tags")
    if resource.get("Attachments"):
        resource_translated["vpc_id"] = [resource.get("Attachments")[0].get("VpcId")]
        resource_translated["attachments"] = resource.get("Attachments")
    return resource_translated


def convert_raw_nat_gateway_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("NatGatewayId")
    describe_parameters = OrderedDict(
        {
            "SubnetId": "subnet_id",
            "ConnectivityType": "connectivity_type",
            "Tags": "tags",
            "State": "state",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in describe_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    if raw_resource.get("NatGatewayAddresses"):
        for nat_gateway_address in raw_resource.get("NatGatewayAddresses"):
            if "AllocationId" in nat_gateway_address:
                resource_translated["allocation_id"] = nat_gateway_address.get(
                    "AllocationId"
                )
                break
    return resource_translated


def convert_raw_elastic_ip_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Given an object state from aws, this function creates a translated resource object in response.

    Args:
        hub: required for functions in hub
        raw_resource (Dict): The dictionary object from where the raw state of resource needs to be translated.
        idem_resource_name (Text): The idem of the idem resource

    Returns: Dict[str, Any]
    """
    elastic_ip = raw_resource
    resource_id = elastic_ip.get("PublicIp")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    describe_parameters = OrderedDict(
        {
            "AllocationId": "allocation_id",
            "Domain": "domain",
            "NetworkBorderGroup": "network_border_group",
            "PublicIpv4Pool": "public_ipv4_pool",
            "Tags": "tags",
            "PublicIp": "resource_id",
            "CustomerOwnedIpv4Pool": "customer_owned_ipv4_pool",
        }
    )
    for parameter_raw, parameter_present in describe_parameters.items():
        if elastic_ip.get(parameter_raw) is not None:
            resource_translated[parameter_present] = elastic_ip.get(parameter_raw)
    return resource_translated



import json
from typing import Dict
from typing import List


def are_lists_identical(hub, list1: List, list2: List) -> bool:
    """
    Compare two lists and logs the difference.
    :param list1: first list.
    :param list2: second list.
    :return: true if there is no difference between both lists.
    """
    if (list1 is None or len(list1) == 0) and (list2 is None or len(list2) == 0):
        return True
    if list1 is None or len(list1) == 0 or list2 is None or len(list2) == 0:
        return False

    diff = [i for i in list1 + list2 if i not in list1 or i not in list2]
    result = len(diff) == 0
    if not result:
        hub.log.debug(f"There are {len(diff)} differences:\n{diff[:5]}")
    return result


def standardise_json(hub, value: str or Dict, sort_keys: bool = True):
    # Format json string or dictionary
    if value is None:
        return None

    if isinstance(value, str):
        json_dict = json.loads(value)
    else:
        json_dict = value
    return json.dumps(json_dict, separators=(", ", ": "), sort_keys=sort_keys)


def is_json_identical(hub, struct1: str, struct2: str):
    return _sorting(json.loads(struct1)) == _sorting(json.loads(struct2))


def _sorting(item):
    # Sort json structure
    # Ignore order in dictionaries and lists
    if isinstance(item, dict):
        return sorted((key, _sorting(values)) for key, values in item.items())
    if isinstance(item, list):
        if all(isinstance(x, str) for x in item):
            return sorted(item)
        else:
            return [_sorting(x) for x in item]
    else:
        return item



"""

Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_security_group
hub.exec.boto3.client.ec2.delete_security_group
hub.exec.boto3.client.ec2.describe_security_groups
hub.tool.boto3.resource.exec(resource, authorize_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, authorize_ingress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_ingress, *args, **kwargs)
"""
import copy
from copy import deepcopy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {"require": ["aws.ec2.vpc.present"]},
}


async def present(
    hub,
    ctx,
    name: str,
    description: str,
    vpc_id: str,
    resource_id: str = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a security group. A security group acts as a virtual firewall for your instance to control inbound and
    outbound traffic. For more information, see Amazon EC2 security groups in the Amazon Elastic Compute Cloud User
    Guide and Security groups for your VPC in the Amazon Virtual Private Cloud User Guide. When you create a
    security group, you specify a friendly name of your choice. You can have a security group for use in EC2-Classic
    with the same name as a security group for use in a VPC. However, you can't have two security groups for use in
    EC2-Classic with the same name or two security groups for use in a VPC with the same name. You have a default
    security group for use in EC2-Classic and a default security group for use in your VPC. If you don't specify a
    security group when you launch an instance, the instance is launched into the appropriate default security
    group. A default security group includes a default rule that grants instances unrestricted network access to
    each other. You can add or remove rules from your security groups using AuthorizeSecurityGroupIngress,
    AuthorizeSecurityGroupEgress, RevokeSecurityGroupIngress, and RevokeSecurityGroupEgress. For more information
    about VPC security group limits, see Amazon VPC Limits.

    Args:
        hub:
        ctx:
        name(Text): Name of the resource.
        description(Text): Description of the security group
        vpc_id(Text): Id of the VPC security group should be attached to
        resource_id(Text, optional): Security Group Id
        tags(Dict, Optional): List of Name, Value pair to tag security group

    Request Syntax:
        [security_group_id]:
          aws.ec2.security_group.present:
          - resource_id: 'string'
          - name: 'string'
          - vpc_id: 'string'
          - tags:
                - Key: 'string'
                  Value: 'string'


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            isolation-securitygroup-84c9a893-844b-40ab-86e8-b565dea88b5d:
              aws.ec2.security_group.present:
                  - resource_id: sg-0008bd25b7867b5cf
                  - name: isolation-securitygroup-84c9a893-844b-40ab-86e8-b565dea88b5d
                  - vpc_id: vpc-247e9a5d
                  - description: sg-description
                  - tags:
                        - Key: Name
                          Value: sg-name
                        - Key: sg-tag-key-2
                          Value: sg-tag-value-2
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        resource = hub.tool.boto3.resource.create(
            ctx, "ec2", "SecurityGroup", resource_id
        )
        before = await hub.tool.boto3.resource.describe(resource)
    if before:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(before)
            plan_state = copy.deepcopy(result["old_state"])
            old_tags = result["old_state"].get("tags")
            resource_updated = False
            if (
                tags is not None
                and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    tags, old_tags
                )
            ):
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx,
                    resource_id=resource_id,
                    old_tags=old_tags,
                    new_tags=tags,
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + (update_ret["comment"])
                    result["result"] = update_ret["result"]
                    return result
                result["result"] = result["result"] and update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
                if ctx.get("test", False) and resource_updated:
                    plan_state["tags"] = update_ret["ret"].get("tags")
                    result["comment"] = result["comment"] + (
                        f"Would update aws.ec2.security_group {name}",
                    )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "description": description,
                    "vpc_id": vpc_id,
                    "tags": tags,
                },
            )
            result["comment"] = result["comment"] + (
                f"Would create aws.ec2.security_group {name}",
            )
            return result
        try:
            ret = await hub.exec.boto3.client.ec2.create_security_group(
                ctx,
                GroupName=name,
                Description=description,
                VpcId=vpc_id,
                TagSpecifications=[{"ResourceType": "security-group", "Tags": tags}]
                if tags is not None
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + (ret["comment"])
                return result
            result["comment"] = result["comment"] + (f"Created '{name}'",)
            resource_id = ret["ret"]["GroupId"]

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            resource = hub.tool.boto3.resource.create(
                ctx, "ec2", "SecurityGroup", resource_id
            )
            after = await hub.tool.boto3.resource.describe(resource)

            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(after)
        else:
            result["new_state"] = deepcopy(result["old_state"])

    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a security group. If you attempt to delete a security group that is associated with an instance, or is
    referenced by another security group, the operation fails with InvalidGroup.InUse in EC2-Classic or
    DependencyViolation in EC2-VPC.

    Args:
        hub:
        ctx:
        name(Text): A name to identify the resource.
        resource_id(Text): Aws Security group ID to identify the resource

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            sg-0008bd25b7867b5cf:
              aws.ec2.security_group.absent:
                - name: isolation-securitygroup-84c9a893-844b-40ab-86e8-b565dea88b5d
                - resource_id: id: sg-0008bd25b7867b5cf
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "SecurityGroup", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = result["comment"] + (
            f"'{name}' is already in deleted state.",
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(before)
        result["comment"] = result["comment"] + (
            f"Would delete aws.ec2.security_group {name}",
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(before)
        try:
            ret = await hub.exec.boto3.client.ec2.delete_security_group(
                ctx, GroupId=resource_id
            )
            if not ret["result"]:
                result["comment"] = result["comment"] + (ret["comment"])
                result["result"] = ret["result"]
                return result
            result["comment"] = result["comment"] + (f"Deleted '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes the specified security groups or all of your security groups. A security group is for use with
    instances either in the EC2-Classic platform or in a specific VPC. For more information, see Amazon EC2 security
    groups in the Amazon Elastic Compute Cloud User Guide and Security groups for your VPC in the Amazon Virtual
    Private Cloud User Guide.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.security_group
    """
    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_security_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe security_group {ret['comment']}")
        return {}

    for security_group in ret["ret"]["SecurityGroups"]:

        translated_security_group = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(security_group)
        )

        result[security_group.get("GroupId")] = {
            "aws.ec2.security_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_security_group.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.eks.create_addon
hub.exec.boto3.client.eks.delete_addon
hub.exec.boto3.client.eks.describe_addon
hub.exec.boto3.client.eks.list_addons
hub.exec.boto3.client.eks.update_addon
"""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.ec2.subnet.absent",
            "aws.ec2.vpc.absent",
            "aws.iam.role.absent",
            "aws.iam.role_policy_attachment.absent",
            "aws.eks.cluster.absent",
        ],
    },
    "present": {
        "require": [
            "aws.ec2.subnet.present",
            "aws.ec2.vpc.present",
            "aws.iam.role.present",
            "aws.iam.role_policy_attachment.present",
            "aws.eks.cluster.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    addon_version: str,
    service_account_role_arn: str = None,
    resolve_conflicts: str = "OVERWRITE",
    tags: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an Amazon EKS add-on. Amazon EKS add-ons help to automate the provisioning and lifecycle management of
    common operational software for Amazon EKS clusters. Amazon EKS add-ons require clusters running version 1.18 or
    later because Amazon EKS add-ons rely on the Server-side Apply Kubernetes feature, which is only available in
    Kubernetes 1.18 and later. For more information, see Amazon EKS add-ons in the Amazon EKS User Guide.

    Args:
        name(Text): A name, ID to identify the resource.
        cluster_name(Text, required): The name of the cluster to create the add-on for
        addon_version(Text, required): The version of the add-on. The version must match one of the versions returned
            by ` DescribeAddonVersions
        service_account_role_arn(Text optional):The Amazon Resource Name (ARN) of an existing IAM role to bind to the add-on's
            service account. The role must be assigned the IAM permissions required by the add-on. If you don't specify
            an existing IAM role, then the add-on uses the permissions assigned to the node IAM role.
        resolve_conflicts(Text optional): How to resolve parameter value conflicts when migrating an existing add-on
            to an Amazon EKS add-on
        tags(Dict, optional): The metadata to apply to the cluster to assist with categorization and organization.
            Each tag consists of a key and an optional value. You define both

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            kube-proxy:
              aws.eks.addon.present:
              - cluster_name: {{ cluster }}
              - addon_version: v1.19.8-eksbuild.1
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    is_not_changed = True
    before = await hub.exec.boto3.client.eks.describe_addon(
        ctx, clusterName=cluster_name, addonName=name
    )
    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.eks.addon {name}"
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.eks.addon {name}"
            result["result"] = True
        return result

    if before:
        result["comment"] = f"'{name}' already exists"
        result["old_state"] = before["ret"]["addon"]
        arn = before["ret"]["addon"]["addonArn"]
        if before["ret"]["addon"].get("clientRequestToken"):
            client_request_token = before["ret"]["addon"]["clientRequestToken"]
        else:
            client_request_token = None
        # change depends on addonVersion, serviceAccountRoleArn
        if (
            addon_version
            and before["ret"]["addon"].get("addonVersion") != addon_version
        ) or (
            service_account_role_arn
            and before["ret"]["addon"].get("serviceAccountRoleArn")
            != service_account_role_arn
        ):
            ret = await hub.exec.boto3.client.eks.update_addon(
                ctx,
                clusterName=cluster_name,
                addonName=name,
                addonVersion=addon_version,
                serviceAccountRoleArn=service_account_role_arn,
                resolveConflicts=resolve_conflicts,
                clientRequestToken=client_request_token,
                tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "eks",
                    "addon_active",
                    None,
                    addonName=name,
                    clusterName=cluster_name,
                )
            except Exception as e:
                result["comment"] = str(e)
                result["result"] = False
                return result
            is_not_changed = False
        # Update tags
        old_tags = (
            copy.deepcopy(before["ret"]["addon"].get("tags"))
            if before["ret"]["addon"].get("tags")
            else {}
        )
        update_tags_ret = await hub.exec.aws.eks.tag.update_eks_tags(
            ctx=ctx,
            resource_arn=arn,
            old_tags=old_tags,
            new_tags=tags,
        )
        if not update_tags_ret["result"]:
            result["comment"] = update_tags_ret["comment"]
            result["result"] = False
            return result
        if update_tags_ret["ret"]:
            is_not_changed = False
        if not is_not_changed:
            result["comment"] = f"Updated '{name}'"

    else:
        try:
            ret = await hub.exec.boto3.client.eks.create_addon(
                ctx,
                clusterName=cluster_name,
                addonName=name,
                addonVersion=addon_version,
                serviceAccountRoleArn=service_account_role_arn,
                resolveConflicts=resolve_conflicts,
                tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            await hub.tool.boto3.client.wait(
                ctx,
                "eks",
                "addon_active",
                None,
                addonName=name,
                clusterName=cluster_name,
            )
            result["comment"] = f"Created '{name}'"
        except Exception as e:
            result["comment"] = str(e)
            result["result"] = False
            return result

    try:
        if is_not_changed:
            after = await hub.exec.boto3.client.eks.describe_addon(
                ctx, clusterName=cluster_name, addonName=name
            )
            result["new_state"] = after["ret"]["addon"]
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    preserve: bool = False,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Delete an Amazon EKS add-on. When you remove the add-on, it will also be deleted from the cluster. You can
    always manually start an add-on on the cluster using the Kubernetes API.

    Args:
        name(Text): A name, ID to identify the resource.
        cluster_name(Text, required): The name of the cluster to delete the add-on from.
        preserve(Boolean, optional): Specifying this option preserves the add-on software on your cluster but Amazon
            EKS stops managing any settings for the add-on. If an IAM account is associated with the add-on,
            it is not removed.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            kube-proxy:
              aws.eks.addon.absent:
              - cluster_name: {{ cluster }}
              - preserve: True
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.eks.describe_addon(
        ctx, clusterName=cluster_name, addonName=name
    )

    if not before:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would delete aws.eks.addon {name}"
        return result
    else:
        result["old_state"] = before["ret"]["addon"]
        try:
            ret = await hub.exec.boto3.client.eks.delete_addon(
                ctx, clusterName=cluster_name, addonName=name, preserve=preserve
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "eks",
                    "addon_deleted",
                    None,
                    addonName=name,
                    clusterName=cluster_name,
                )
            except Exception as e:
                result["comment"] = str(e)
                result["result"] = False
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the available add-ons.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.eks.addon
    """

    result = {}
    cluster_ret = await hub.exec.boto3.client.eks.list_clusters(ctx)
    if not cluster_ret["result"]:
        hub.log.debug(f"Could not describe cluster {cluster_ret['comment']}")
        return {}
    cluster_addons = []
    try:
        for name in cluster_ret["ret"]["clusters"]:
            addon_ret = await hub.exec.boto3.client.eks.list_addons(
                ctx, clusterName=name
            )
            if not addon_ret["result"]:
                result["comment"] = addon_ret["comment"]
                return result
            for addon in addon_ret["ret"]["addons"]:
                cluster_addons.append({"name": name, "addon": addon})
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
        return result

    describe_parameters = OrderedDict(
        {
            "clusterName": "cluster_name",
            "addonArn": "addon_arn",
            "addonVersion": "addon_version",
            "releaseVersion": "release_version",
            "status": "status",
            "serviceAccountRoleArn": "service_account_role_arn",
        }
    )

    for cluster_addon in cluster_addons:
        cluster_name = cluster_addon["name"]
        current_addon = cluster_addon["addon"]
        describe_ret = await hub.exec.boto3.client.eks.describe_addon(
            ctx, clusterName=cluster_name, addonName=current_addon
        )
        add_on = describe_ret["ret"]["addon"]
        new_addon = []
        for parameter_old_key, parameter_new_key in describe_parameters.items():
            if add_on.get(parameter_old_key) is not None:
                new_addon.append({parameter_new_key: add_on.get(parameter_old_key)})
        if add_on.get("tags"):
            new_addon.append({"tags": add_on.get("tags").copy()})

        result[current_addon] = {"aws.eks.addon.present": new_addon}
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS Elasticache to present input format.
"""


async def convert_raw_elasticache_subnet_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("CacheSubnetGroupName")
    resource_parameters = OrderedDict(
        {
            "CacheSubnetGroupDescription": "cache_subnet_group_description",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("Subnets"):
        subnet_ids_list = []
        for subnets in raw_resource.get("Subnets"):
            if "SubnetIdentifier" in subnets:
                subnet_ids_list.append(subnets.get("SubnetIdentifier"))
        resource_translated["subnet_ids"] = subnet_ids_list
    if raw_resource.get("ARN"):
        tags = await hub.exec.boto3.client.elasticache.list_tags_for_resource(
            ctx, ResourceName=raw_resource.get("ARN")
        )
        if tags["result"]:
            if tags["ret"]["TagList"]:
                resource_translated["tags"] = tags["ret"]["TagList"]
    return resource_translated



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_role_tags(
    hub,
    ctx,
    role_name: str,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
) -> Dict[str, Any]:
    """
    Update tags of AWS IAM Role

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        role_name: AWS IAM role name
        old_tags: list of old tags
        new_tags: list of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}
    """
    tags_to_add = []
    old_tags_map = {tag.get("Key"): tag for tag in old_tags}
    tags_result = copy.deepcopy(old_tags_map)
    for tag in new_tags:
        if tag.get("Key") in old_tags_map:
            if tag.get("Value") == old_tags_map.get(tag.get("Key")).get("Value"):
                del old_tags_map[tag.get("Key")]
            else:
                tags_to_add.append(tag)
        else:
            tags_to_add.append(tag)
    tags_to_remove = [tag.get("Key") for tag in old_tags_map.values()]
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.iam.untag_role(
                ctx, RoleName=role_name, TagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.iam.tag_role(
                ctx, RoleName=role_name, Tags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",)
    return result


async def update_role(
    hub,
    ctx,
    old_state: Dict[str, Any],
    description: str = None,
    max_session_duration: int = None,
):
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}
    if (description is not None) and old_state.get("description") != description:
        update_payload["Description"] = description
    if (max_session_duration is not None) and old_state.get(
        "max_session_duration"
    ) != max_session_duration:
        update_payload["MaxSessionDuration"] = max_session_duration
    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.iam.update_role(
                ctx=ctx, RoleName=old_state.get("name"), **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        if "MaxSessionDuration" in update_payload:
            result["ret"]["max_session_duration"] = update_payload["MaxSessionDuration"]
            result["comment"] = result["comment"] + (
                f"Update max_session_duration: {update_payload['MaxSessionDuration']}",
            )
        if "Description" in update_payload:
            result["ret"]["description"] = update_payload["Description"]
            result["comment"] = result["comment"] + (
                f"Update description: {update_payload['Description']}",
            )
    return result


async def update_policy(
    hub,
    ctx,
    role_name: str,
    policy: str,
):
    # Update role's trust policy (the policy that is embedded within the role)
    result = dict(comment=(), result=True, ret=None)

    if not ctx.get("test", False):
        update_ret = await hub.exec.boto3.client.iam.update_assume_role_policy(
            ctx=ctx, RoleName=role_name, PolicyDocument=policy
        )
        if not update_ret["result"]:
            result["comment"] = update_ret["comment"]
            result["result"] = False
            return result

    if ctx.get("test", False):
        result["comment"] = (f"Would update aws.iam.role '{role_name}' policy",)
    else:
        result["comment"] = (f"Updated aws.iam.role '{role_name}' policy",)
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

"""
Util functions to convert raw resource state from AWS EC2 to present input format.
"""


def convert_raw_db_subnet_group_to_present(
    hub, resource: Dict[str, Any], tags: List = None, idem_resource_name: str = None
) -> Dict[str, Any]:
    new_resource = {}
    db_subnet_group_id = resource.get("DBSubnetGroupName")
    describe_parameters = OrderedDict(
        {
            "DBSubnetGroupDescription": "db_subnet_group_description",
            "DBSubnetGroupArn": "db_subnet_group_arn",
        }
    )
    new_resource = {"name": db_subnet_group_id, "resource_id": db_subnet_group_id}
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if resource.get(parameter_old_key) is not None:
            new_resource[parameter_new_key] = resource.get(parameter_old_key)
    if resource.get("Subnets"):
        subnets = []
        for each_subnet in resource.get("Subnets"):
            subnets.append(each_subnet.get("SubnetIdentifier"))

        new_resource["subnets"] = subnets
    if tags:
        new_resource["tags"] = tags
    return new_resource



"""
Plugin for aws enforced state management
"""
import json
import pathlib
from typing import Any
from typing import Dict

import botocore.exceptions


def __init__(hub):
    hub.esm.aws.ACCT = ["aws"]


async def get_state(hub, ctx) -> Dict[str, Any]:
    """
    Read this context's state from AWS S3 bucket. Return empty dict if the file key path does not exist in the bucket.

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.

    Returns:
        A dict containing all resource state records. Return empty dict if the file key path does not exist in the bucket.
    """
    esm = ctx.acct.get("esm")
    try:
        client = hub.tool.boto3.client.get_client(ctx, service_name="s3")
        content = client.get_object(Bucket=esm.get("bucket"), Key=esm.get("key"))
        state = json.loads(content["Body"].read())
        return state
    except botocore.exceptions.ClientError as client_ex:
        if client_ex.response["Error"]["Code"] == "NoSuchKey":
            hub.log.debug(
                f"No state file found from S3 bucket {esm.get('bucket')} using key {esm.get('key')}"
            )
            return {}
        else:
            hub.log.error(
                f"Failed to retrieve state file from S3 bucket {esm.get('bucket')} using key {esm.get('key')}"
                f" with error {client_ex.response['Error']['Message']}"
            )
            raise client_ex
    except Exception as e:
        hub.log.error(
            f"Failed to retrieve state file from S3 bucket {esm.get('bucket')} using key {esm.get('key')}"
            f" with error {str(e)}"
        )
        raise e


async def set_state(hub, ctx, state: Dict[str, Any]):
    """
    Write the state to this context's cache_file

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        state: A dict containing all resource states that will be stored in S3 bucket.

    Returns:
        None if the operation succeeds. An exception will raise if the operation failed.
    """
    try:
        esm = ctx.acct.get("esm")
        ret = await hub.exec.boto3.client.s3.put_object(
            ctx,
            Bucket=esm.get("bucket"),
            Body=bytes(json.dumps(state), "utf-8"),
            Key=esm.get("key"),
        )
        if not ret["result"]:
            error_msg = f"Failed to upload state file to S3 bucket {esm.get('bucket')} as {esm.get('key')} with error {ret['comment']}"
            raise RuntimeError(error_msg)
        else:
            hub.log.debug(
                f"Successfully upload state file to S3 bucket {esm.get('bucket')} as {esm.get('key')}"
            )
            return
    except Exception as e:
        hub.log.error(str(e))
        raise e


async def enter(hub, ctx):
    """
    Verify that only one process can access the same enforced state record file in S3 bueckt.
    Set a lock by performing an atomic operation to put the file key to DynamoDB.

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.

    Returns:
        None if lock has been acquired successfully. Raise exception if lock was failed to be acquired
    """
    esm = ctx.acct.get("esm")
    if esm is None:
        raise ValueError("ESM profile is required to support remote ESM storage.")
    if not esm.get("dynamodb_table"):
        raise ValueError("dynamodb_table is required in ESM profile.")
    if not esm.get("bucket"):
        raise ValueError("bucket is required in ESM profile.")
    if not esm.get("key"):
        raise ValueError("key is required in ESM profile.")
    try:
        # Put item with conditional expression to acquire the lock. The conditional check and put operation is atomic.
        # This ensures that the S3 bucket file can be locked properly.
        db_resource = hub.tool.boto3.resource.create(
            ctx, "dynamodb", "Table", esm.get("dynamodb_table")
        )
        db_resource.put_item(
            Item={"LockID": f"{esm.get('bucket')}/{esm.get('key')}"},
            ConditionExpression="attribute_not_exists(#r)",
            ExpressionAttributeNames={"#r": "LockID"},
        )
        # Lock acquired
        return
    except botocore.exceptions.ClientError as e:
        # Another exception than ConditionalCheckFailedException was caught, raise as-is
        if e.response["Error"]["Code"] != "ConditionalCheckFailedException":
            raise e
        else:
            # If conditional check failed, it means the lock cannot be acquired because it is already locked
            raise ValueError(
                f"File {esm.get('bucket')}/{esm.get('key')} has been locked by another user/process."
            )


async def exit_(hub, ctx, handle: pathlib.Path, exception: Exception):
    """
    Release the db lock and handle errors

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        handle: This parameter is inherited from Idem. It is not being used in Idem-aws plugin.
        exception: Error.

    Returns:
        None if file lock has been released. Raise Exception if any failure happened.
    """
    if exception:
        hub.log.error(f"{exception.__class__.__name__}: {exception}")
    esm = ctx.acct.get("esm")
    if esm is None:
        raise ValueError("ESM profile is required to support remote ESM storage.")
    try:
        # Delete item with conditional expression to release the lock
        db_resource = hub.tool.boto3.resource.create(
            ctx, "dynamodb", "Table", esm.get("dynamodb_table")
        )
        db_resource.delete_item(
            Key={"LockID": f"{esm.get('bucket')}/{esm.get('key')}"},
            ConditionExpression="attribute_exists(#r)",
            ExpressionAttributeNames={"#r": "LockID"},
        )
        # Lock released
        return
    except botocore.exceptions.ClientError as e:
        # Another exception than ConditionalCheckFailedException was caught, raise as-is
        if e.response["Error"]["Code"] != "ConditionalCheckFailedException":
            raise e
        else:
            # Else, lock has been released
            hub.log.debug(
                f"Lock {esm.get('bucket')}/{esm.get('key')} has been previously released."
            )
            return



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_user_tags(
    hub,
    ctx,
    user_name: str,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
) -> Dict[str, Any]:
    """
    Update tags of AWS IAM user

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        user_name: AWS IAM username
        old_tags: list of old tags
        new_tags: list of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}
    """
    tags_to_add = []
    old_tags_map = {tag.get("Key"): tag for tag in old_tags}
    tags_result = copy.deepcopy(old_tags_map)
    for tag in new_tags:
        if tag.get("Key") in old_tags_map:
            if tag.get("Value") == old_tags_map.get(tag.get("Key")).get("Value"):
                del old_tags_map[tag.get("Key")]
            else:
                tags_to_add.append(tag)
        else:
            tags_to_add.append(tag)
    tags_to_remove = [tag.get("Key") for tag in old_tags_map.values()]
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.iam.untag_user(
                ctx, UserName=user_name, TagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.iam.tag_user(
                ctx, UserName=user_name, Tags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",)
    return result


async def update_user(
    hub,
    ctx,
    before: Dict[str, Any],
    user_name: str = None,
    path: str = None,
):
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}
    if user_name and before.get("user_name") != user_name:
        update_payload["NewUserName"] = user_name
    if path and before.get("path") != path:
        update_payload["NewPath"] = path
    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.iam.update_user(
                ctx=ctx, UserName=before.get("resource_id"), **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        if "NewUserName" in update_payload:
            result["ret"]["user_name"] = update_payload["NewUserName"]
            result["comment"] = result["comment"] + (
                f"Update user_name: {update_payload['NewUserName']} on user {before.get('user_name')}",
            )
        if "NewPath" in update_payload:
            result["ret"]["path"] = update_payload["NewPath"]
            result["comment"] = result["comment"] + (
                f"Update path: {update_payload['NewPath']} on user {before.get('user_name')}",
            )
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """
    Update tags of AWS Dynamo DB resources

    Args:
        resource_arn: Identifies the Amazon DynamoDB resource to which tags should be added.
            This value is an Amazon Resource Name (ARN).
        old_tags: List of existing tags
        new_tags: List of new tags

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": None}

    """

    tags_to_add = []
    tags_to_remove = []
    old_tags_map = {tag.get("Key"): tag for tag in old_tags or []}
    tags_result = copy.deepcopy(old_tags_map)
    if new_tags is not None:
        for tag in new_tags:
            if tag.get("Key") in old_tags_map:
                if tag.get("Value") == old_tags_map.get(tag.get("Key")).get("Value"):
                    del old_tags_map[tag.get("Key")]
                else:
                    tags_to_add.append(tag)
            else:
                tags_to_add.append(tag)
        tags_to_remove = list(old_tags_map.keys())
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        return result

    if tags_to_remove:
        delete_error_message = f"Could not delete tags {tags_to_remove} for DynamoDB table with ARN {resource_arn}"
        if not ctx.get("test", False):
            delete_tag_resp = await hub.exec.boto3.client.dynamodb.untag_resource(
                ctx, ResourceArn=resource_arn, TagKeys=tags_to_remove
            )
            if not delete_tag_resp:
                hub.log.debug(delete_error_message)
                result["comment"] = (f"Could not delete tags {tags_to_remove}",)
                result["result"] = False
                return result
        [tags_result.pop(key) for key in tags_to_remove]
    if tags_to_add:
        add_error_message = f"Could not create tags {tags_to_remove} for DynamoDB table with ARN {resource_arn}"
        if not ctx.get("test", False):
            create_tag_resp = await hub.exec.boto3.client.dynamodb.tag_resource(
                ctx, ResourceArn=resource_arn, Tags=tags_to_add
            )
            if not create_tag_resp:
                hub.log.debug(add_error_message)
                result["comment"] = result["comment"] + (
                    f"Could not update tags {tags_to_add}",
                )
                result["result"] = False
                return result

    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    result["comment"] = result["comment"] + (
        f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",
    )
    return result
